# Legacy Optimization Passes
from collections import defaultdict
from typing import Callable, Dict, Iterable, List

import numpy as np
import torch
from tqdm import tqdm

from esp_ppq.core import *
from esp_ppq.core import QuantizationStates
from esp_ppq.executor import BaseGraphExecutor, TorchExecutor
from esp_ppq.IR import BaseGraph, Operation, QuantableOperation, SearchableGraph
from esp_ppq.IR.quantize import QuantableVariable
from esp_ppq.IR.search import TraversalCommand
from esp_ppq.quantization.algorithm.training import *
from esp_ppq.quantization.measure import torch_mean_square_error

from .base import QuantizationOptimizationPass
from .training import TrainingBasedPass


class TimeDecay:
    """A helper class computing time decay."""

    def __init__(self, t_max: int, decay: float = 0.2, beta_start: float = 20, beta_end: float = 2):
        self.t_max = t_max
        self.start_decay = decay * t_max
        self.start_b = beta_start
        self.end_b = beta_end

    def __call__(self, t):
        rel_t = (t - self.start_decay) / (self.t_max - self.start_decay)
        return self.end_b + 0.5 * (self.start_b - self.end_b) * (1 + np.cos(rel_t * np.pi))


class AdaroundRegTerm(torch.nn.Module):
    """Adaround Reg Term is a part of Adaround optimization algorithm.
    This term represents the difference between a fp32 value and its quantized counter-part.
        We use a same implementation as proposed in Adaround paper.
    Args:
        torch ([type]): [description]
    """

    def __init__(
        self,
        max_iter: int = 20000,
        zeta: float = 1.1,
        gamma: float = -0.1,
        alpha: float = 0.01,
        beta: float = 20,
        warm_ratio: float = 0.2,
    ):
        self.max_iter = max_iter
        self.zeta = zeta
        self.gamma = gamma
        self.alpha = alpha
        self.beta = beta
        self.warm_ratio = warm_ratio
        self.temp_anneal = TimeDecay(self.max_iter, self.warm_ratio)
        super().__init__()

    def rectified_sigmoid(self, r: torch.Tensor) -> torch.Tensor:
        return ((self.zeta - self.gamma) * torch.sigmoid(r) + self.gamma).clamp(0, 1)

    def forward(self, r: torch.Tensor, iter: int) -> torch.Tensor:
        if iter < self.max_iter * self.warm_ratio:
            round_loss = 0
        else:
            self.beta = self.temp_anneal(iter)
            round_loss = self.alpha * (1 - torch.pow((self.rectified_sigmoid(r) - 0.5).abs() * 2, self.beta)).sum()
        return round_loss


class AdaRoundDelegator(TorchQuantizeDelegator):
    def __init__(
        self,
        var: QuantableVariable,
        config: TensorQuantizationConfig,
        steps: int,
    ) -> None:
        self.reg = AdaroundRegTerm(max_iter=steps)
        self.config = config
        self.var = var
        self.is_parameter = self.var.is_parameter
        self.rounding = self.initiate_rounding(value=self.var.value, config=self.config, zeta=1.1, gamma=-0.1)

        if not self.var.is_parameter:
            raise TypeError(
                f'Can not create adaround delegator with variable {var.name}, '
                'Adaround delegator works only with parameter.'
            )
        if self.config.state == QuantizationStates.PASSIVE:
            raise TypeError(
                f'Can not create adaround delegator with variable {var.name}, '
                'Adaround delegator can not work with passive parameter.'
            )
        self.param_backup = None
        if self.is_parameter:
            self.param_backup = self.var.value.clone()

    @staticmethod
    def initiate_rounding(
        value: torch.Tensor, config: TensorQuantizationConfig, zeta: float, gamma: float
    ) -> torch.Tensor:
        with torch.no_grad():
            scale, offset = config.scale, config.offset
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                shape = [1 if axis != config.channel_axis else -1 for axis in range(value.ndim)]
                scale = scale.view(shape)

            rounding = (value / scale) - (value / scale).floor()
            rounding = -torch.log((zeta - gamma) / (rounding - gamma) - 1)
            rounding = torch.zeros_like(rounding).copy_(rounding)
            rounding.requires_grad = True
        return rounding

    def trainable_tensors(self) -> List[torch.Tensor]:
        tensors = [self.rounding]
        return tensors

    def finalize(self) -> None:
        weight, scale, offset = self.var.value, self.config.scale, self.config.offset
        if self.config.policy.has_property(QuantizationProperty.PER_CHANNEL):
            shape = [1 if axis != self.config.channel_axis else -1 for axis in range(weight.ndim)]
            scale = scale.view(shape)
            offset = offset.view(shape)
        weight = (weight / scale).floor() + (self.rounding >= 0).float()
        weight = torch.clamp(weight + offset, self.config.quant_min, self.config.quant_max)
        weight = (weight - offset) * scale
        self.var.value = weight

    def withdraw(self) -> None:
        with torch.no_grad():
            self.var.value.copy_(self.param_backup)

    def __call__(self, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        scale = config.scale
        offset = config.offset
        if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
            shape = [1 if axis != config.channel_axis else -1 for axis in range(tensor.ndim)]
            scale = scale.view(shape)
            offset = offset.view(shape)
        tensor = (tensor / scale).floor() + self.reg.rectified_sigmoid(self.rounding)
        tensor = torch.clamp(tensor + offset, config.quant_min, config.quant_max)
        tensor = (tensor - offset) * scale
        return tensor

    def regularization_loss(self, step: int) -> torch.Tensor:
        return self.reg.forward(r=self.rounding, iter=step)


class AdaroundPass(TrainingBasedPass):
    """Blockwise Reconstruction Pass, perform adaround block by block, if you
    specify interested_layers in the setting, then only block which contains
    any of operations specified in interested_layers will be optimized,
    otherwise all searched blocks will be optimized. A standard procedure is,
    first turn all training-based optimization passes off in your quantization
    setting and run a plain quantization, then use error analysis tool(provided
    by ppq) to analysis snr error or cosine similarities of every layer, choose
    names of those with significant snr or poor similarities as your
    interested_layers, then turn on this pass and do optimization. In case you
    have no idea which layers should be selected as interested_layers, simply
    leave it as blank and all blocks will be tuned. Note that you could control
    the maximum number of operations in a block by setting
    OPTIM_ADVOPT_GRAPH_MAXSIZE in esp_ppq.core.common, and by default every block
    will be trained for 300 epochs, which takes certain long time. The
    optimization goal of every block is.
                Loss = LpNormLoss(y, y^) + lambda * rounding_loss(v)
    where y is the output of the current block running in fp32 mode, and y^ is the output of the current block running
    in quant mode, lambda is a hyperparameter adjusting scales of rounding loss, and v is the element-wise rounding
    parameter applied to weights of every computing op in the block.
    """

    def __init__(
        self,
        name: str = 'Block-wise Adaround Reconstruction',
        interested_layers: List[str] = [],
        is_scale_trainable: bool = False,
        steps: int = 8000,
        lr: float = 1e-3,
        gamma: float = 1.0,
        collecting_device: str = 'cuda',
        block_size: int = 4,
    ) -> None:
        super().__init__(name=name)
        self.interested_layers = interested_layers
        self.lr = lr
        self.gamma = gamma
        self.steps = steps
        self.block_size = block_size
        self.collecting_device = collecting_device
        self.is_scale_trainable = is_scale_trainable
        self.loss_fn = torch_mean_square_error

    def tune_block_weight_scale(
        self, block: TrainableBlock, steps: int = 900, loss_fn: Callable = torch_mean_square_error
    ) -> None:
        # before we tune weight roundings and activation scales, we optimize weight scale by
        # minimizing MSE(W, W^), 900 epochs would be enough in this non-overfit setting. Note
        # that this is usually unnecessary in 8 bit quantization, but we do it it anyway and
        # the loss checking procedure makes sure we always obtain no worse results.
        for op in block.rps:
            if op.is_computing_op and isinstance(op, QuantableOperation):
                c, v = op.input_quant_config[1], op.inputs[1]
                delegator = LSQDelegator(config=c, var=v, is_parameter_trainable=False)
                params = delegator.trainable_tensors()
                if len(params) == 0:
                    continue

                optimizer = torch.optim.Adam(params, lr=self.lr)
                scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, [int(steps / 2), int(steps * 2 / 3)])

                initial_loss = loss_fn(delegator(tensor=v.value, config=c), v.value)
                for _ in range(steps):
                    optimizer.zero_grad()
                    loss = loss_fn(delegator(tensor=v.value, config=c), v.value)
                    loss.backward()
                    optimizer.step()
                    scheduler.step()
                post_loss = loss_fn(delegator(tensor=v.value, config=c), v.value)

                if post_loss > initial_loss:
                    delegator.withdraw()

    def finetune(
        self,
        steps: int,
        learning_rate: float,
        block: TrainableBlock,
        executor: TorchExecutor,
        qt_inputs: List[Dict[str, torch.Tensor]],
        fp_outputs: List[Dict[str, torch.Tensor]],
        optimizer: torch.optim.Optimizer = None,
        scheduler: object = None,
    ) -> Tuple[float, float]:
        # step - 1: enable gradient for training.
        self.enable_block_gradient(block)

        # record pre training loss.
        pre_loss = self.compute_block_loss(
            block=block, qt_inputs=qt_inputs, fp_outputs=fp_outputs, executor=executor, loss_fn=self.loss_fn
        )

        # tune block weight scale
        self.tune_block_weight_scale(block=block)

        # collect trainable params
        trainable_params, delegators = [], {}
        for op in block.rps:
            if not isinstance(op, QuantableOperation):
                continue

            # register quant delegator
            for cfg, var in op.config_with_variable:
                if cfg.state not in {QuantizationStates.ACTIVATED, QuantizationStates.PASSIVE}:
                    continue
                if var.is_parameter and cfg.state not in {QuantizationStates.PASSIVE}:
                    delegator = AdaRoundDelegator(config=cfg, var=var, steps=steps)
                    trainable_params.extend(delegator.trainable_tensors())
                    executor.register_quantize_delegate(config=cfg, delegator=delegator)
                    delegators[cfg] = delegator
                elif self.is_scale_trainable:
                    delegator = LSQDelegator(config=cfg, var=var, is_offset_trainable=False)
                    trainable_params.extend(delegator.trainable_tensors())
                    executor.register_quantize_delegate(config=cfg, delegator=delegator)
                    delegators[cfg] = delegator

        # check if empty.
        tensors = [tensor for tensor in trainable_params if tensor.requires_grad]
        if len(tensors) == 0:
            for cfg, delegator in delegators.items():
                executor.remove_quantize_delegate(config=cfg)
            return 0, 0

        # initilize optimizer.
        if optimizer is None:
            optimizer = torch.optim.Adam(tensors, lr=learning_rate)

        dataset_length = len(qt_inputs)
        if dataset_length == 0:
            raise ValueError('Dataset is empty.')

        # step 2 - training procedure
        for idx in tqdm(range(steps), desc='# Tuning Procedure '):
            qt_input, fp_output = qt_inputs[idx % dataset_length], fp_outputs[idx % dataset_length]

            # forward
            optimizer.zero_grad()
            feed_dict = {k: v.to(executor._device) for k, v in qt_input.items()}
            output_names = [name for name in fp_output]

            qt_output = executor.partial_graph_forward(
                operations=block.rps, feed_dict=feed_dict, output_names=output_names
            )

            # compute loss
            loss = 0.0
            for idx, name in enumerate(output_names):
                loss += self.loss_fn(qt_output[idx], fp_output[name].to(executor._device))

            # collect reg terms
            for delegator in delegators.values():
                if isinstance(delegator, AdaRoundDelegator):
                    loss += delegator.regularization_loss(idx) * self.gamma

            # backward from loss
            assert isinstance(loss, torch.Tensor)
            loss.backward()
            optimizer.step()
            if scheduler is not None:
                scheduler.step()

        # step - 3: record post training loss
        post_loss = self.compute_block_loss(
            block=block, qt_inputs=qt_inputs, fp_outputs=fp_outputs, executor=executor, loss_fn=self.loss_fn
        )

        # check and withdraw
        for cfg, delegator in delegators.items():
            if post_loss > pre_loss:
                delegator.withdraw()
            else:
                delegator.finalize()
            executor.remove_quantize_delegate(config=cfg)

        # disable gradient for evaluation.
        self.disable_block_gradient(block)

        # clear cache
        torch.cuda.empty_cache()
        return pre_loss, post_loss

    def optimize(
        self, graph: BaseGraph, dataloader: Iterable, executor: TorchExecutor, collate_fn: Callable, **kwargs
    ) -> None:
        blocks = self.split_graph_into_blocks(
            graph=graph,
            executing_order=executor._executing_order,
            blocksize=self.block_size,
            interested_layers=self.interested_layers,
        )

        # do per-block finetune
        for block_idx, block in enumerate(blocks):
            # collect data for training
            qt_inputs, fp_outputs = self.collect(
                graph=graph,
                block=block,
                executor=executor,
                dataloader=dataloader,
                collate_fn=collate_fn,
                collecting_device=self.collecting_device,
            )

            print(f'# Block [{block_idx + 1} / {len(blocks)}]: [{block.sp.name} -> {block.ep.name}]')
            pre_loss, post_loss = self.finetune(
                steps=self.steps,
                learning_rate=self.lr,
                block=block,
                qt_inputs=qt_inputs,
                fp_outputs=fp_outputs,
                executor=executor,
            )
            print(f'# Tuning Finished  : ({pre_loss:.4f} -> {min(pre_loss, post_loss):.4f}) [Block Loss]')
            print('')  # blank line


class PPLCudaAddConvReluMerge(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__(name='PPL CUDA Conv(Relu) - Add - Relu Merge')

    def is_same_platform(self, operations: List[Operation]):
        platforms = [operation.platform for operation in operations]
        return all([platform == platforms[0] for platform in platforms])

    def optimize(self, processor: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        def ep_expr(operation: Operation):
            if not isinstance(operation, QuantableOperation):
                return False
            if operation.type == 'Conv':
                return True
            if operation.type in PPLCUDA_ACTIVATIONS:
                upstream_ops = graph.get_upstream_operations(operation=operation)
                if len(upstream_ops) == 0 and upstream_ops[0].type == 'Conv':
                    return True
                if upstream_ops[0] in merged:
                    return True
            return False

        def retrospect(operation: QuantableOperation) -> QuantableOperation:
            if not isinstance(operation, QuantableOperation):
                return None
            if len(graph.get_upstream_operations(operation)) != 1:
                return None

            parent = graph.get_upstream_operations(operation)[0]
            if parent.type != 'Conv':
                return None
            if not isinstance(parent, QuantableOperation):
                return None
            return parent

        def merge_fn(operation: QuantableOperation):
            assert isinstance(operation, QuantableOperation) and operation.type == 'Add'
            # check if upstream ops can be merged
            up_ops = graph.get_upstream_operations(operation)
            if not self.is_same_platform(up_ops + [operation]):
                return

            # Conv - Add - Relu Merge
            config = operation.config.output_quantization_config[0]

            # Step - 1: merge add output to next activation.
            down_ops = graph.get_downstream_operations(operation)
            if (
                len(down_ops) == 1
                and down_ops[0].type in PPLCUDA_ACTIVATIONS
                and isinstance(down_ops[0], QuantableOperation)
                and down_ops[0].platform == operation.platform
            ):
                config.dominated_by = down_ops[0].config.output_quantization_config[0]

            # Step - 2: disable input conv's quantization(only one).
            up_ops = graph.get_upstream_operations(operation)
            assert len(up_ops) == 2, f'Opeartion {operation.name} should has exact 2 input operations.'

            target_operation = None
            for op in up_ops:
                if op.type == 'Conv':
                    target_operation = op
                elif op.type in PPLCUDA_ACTIVATIONS:
                    target_operation = retrospect(operation)
                if target_operation is not None:
                    break

            if target_operation is not None:
                target_operation.config.output_quantization_config[0].dominated_by = config

        graph, merged, unchanged = processor.graph, set(), False

        # merge conv - add iteratively, until there is no one left.
        while not unchanged:
            unchanged = True

            search_engine = SearchableGraph(processor)
            matchings = search_engine(
                TraversalCommand(
                    sp_expr=lambda x: (x.type == 'Add' and isinstance(x, QuantableOperation) and x not in merged),
                    rp_expr=lambda x, y: False,
                    ep_expr=ep_expr,
                    direction='up',
                )
            )

            # count how many matched inputs does an add operation has.
            counter = defaultdict(lambda: 0)

            # path[0] is add operation.
            for path in matchings:
                counter[path[0]] += 1

            for operation, count in counter.items():
                if count == 2:
                    merge_fn(operation)
                    merged.add(operation)
                    unchanged = False



# Test Quantization System Performace on MMSegmentation models with CityScapes

# Quantizer Configuration
SYMMETRICAL = True
PERCHANNEL = True
POWER_OF_2 = False
BIT_WIDTH = 8
FP8 = True

# Should contains model file(.onnx)
MODEL_DIR = 'QuantZoo/Model/mmseg'

# Should contains Calib & Test Img Folder
CALIB_DIR = 'QuantZoo/Data/Cityscapes/Calib'
TEST_DIR = 'QuantZoo/Data/Cityscapes/Test'

# calibration & test batchsize
BATCHSIZE = 1

# write report to here
REPORT_DIR = 'QuantZoo/Reports'

CONFIGS = [
    {'Model': 'stdc1_512x1024_80k_cityscapes', 'Output': ['/convs/convs.0/activate/Relu_output_0']},
    {'Model': 'pspnet_r50-d8_512x1024_40k_cityscapes', 'Output': ['/bottleneck/activate/Relu_output_0']},
    {
        'Model': 'pointrend_r50_512x1024_80k_cityscapes',  # complex model
        'Output': ['/Concat_60_output_0'],
    },
    {'Model': 'fpn_r50_512x1024_80k_cityscapes', 'Output': ['/Add_2_output_0']},
    {'Model': 'icnet_r18-d8_832x832_80k_cityscapes', 'Output': ['/convs/convs.0/activate/Relu_output_0']},
    {'Model': 'fcn_unet_s5-d16_4x4_512x1024_160k_cityscapes', 'Output': ['/convs/convs.0/activate/Relu_output_0']},
    {
        'Model': 'fast_scnn_lr0.12_8x4_160k_cityscapes',
        'Output': ['/convs/convs.1/pointwise_conv/activate/Relu_output_0'],
    },
    {'Model': 'fcn_r50-d8_512x1024_40k_cityscapes', 'Output': ['/conv_cat/activate/Relu_output_0']},
    {'Model': 'bisenetv2_fcn_4x4_1024x1024_160k_cityscapes', 'Output': ['/convs/convs.0/activate/Relu_output_0']},
    {'Model': 'deeplabv3_r50-d8_512x1024_40k_cityscapes', 'Output': ['/bottleneck/activate/Relu_output_0']},
]

import os

import torch
from QuantZoo.Data.Cityscapes.Data import load_cityscapes_dataset
from QuantZoo.Data.Cityscapes.Eval import evaluation
from QuantZoo.Quantizers import MyFP8Quantizer, MyInt8Quantizer
from QuantZoo.Util import error_analyze, report

import esp_ppq.lib as PFL
from esp_ppq.api import ENABLE_CUDA_KERNEL, load_onnx_graph
from esp_ppq.core import TargetPlatform
from esp_ppq.executor import TorchExecutor
from esp_ppq.quantization.optim import (
    LayerwiseEqualizationPass,
    LearnedStepSizePass,
    ParameterQuantizePass,
    RuntimeCalibrationPass,
)

calib_loader = load_cityscapes_dataset(img_folder=CALIB_DIR)
test_loader = load_cityscapes_dataset(img_folder=TEST_DIR)


with ENABLE_CUDA_KERNEL():
    for config in CONFIGS:
        model = config['Model']
        monitoring_vars = config['Output']

        print(f"Ready to run quant benchmark on {model}")
        graph = load_onnx_graph(onnx_import_file=os.path.join(MODEL_DIR, model + '.onnx'))
        quantizer = MyInt8Quantizer(
            graph=graph, sym=SYMMETRICAL, power_of_2=POWER_OF_2, num_of_bits=BIT_WIDTH, per_channel=PERCHANNEL
        )
        if FP8:
            quantizer = MyFP8Quantizer(graph=graph, calibration='floating')

        # convert op to quantable-op
        for name, op in graph.operations.items():
            if op.type in {'Conv', 'ConvTranspose', 'MatMul', 'Gemm'}:
                quantizer.quantize_operation(name, platform=TargetPlatform.INT8)

        # build quant pipeline.
        pipeline = PFL.Pipeline(
            [
                # LayerwiseEqualizationPass(iteration=10),
                ParameterQuantizePass(),
                RuntimeCalibrationPass(),
                # LearnedStepSizePass(steps=500, collecting_device='cpu')
            ]
        )

        # call pipeline.
        executor = TorchExecutor(graph=graph)
        executor.tracing_operation_meta(torch.zeros(size=[BATCHSIZE, 3, 1024, 2048]).cuda())

        pipeline.optimize(
            graph=graph,
            dataloader=calib_loader,
            verbose=True,
            calib_steps=32,
            collate_fn=lambda x: x[0].to('cuda'),
            executor=executor,
        )

        miou = evaluation(graph=graph, dataloader=test_loader, working_directory=TEST_DIR)
        print(f'Model Performance on CityScapes Miou: {miou * 100: .4f}%')

        # error analyze
        performance = error_analyze(
            graph=graph, outputs=monitoring_vars, dataloader=test_loader, collate_fn=lambda x: x[0].cuda(), verbose=True
        )



from collections import defaultdict
from typing import Callable, Dict, Iterable, List

import torch
from tqdm import tqdm

from esp_ppq.core import empty_ppq_cache
from esp_ppq.executor import BaseGraphExecutor
from esp_ppq.executor.torch import TorchExecutor
from esp_ppq.IR import Operation, QuantableOperation, SearchableGraph, TraversalCommand
from esp_ppq.IR.base.graph import BaseGraph
from esp_ppq.quantization.algorithm.equalization import EqualizationPair

from .base import QuantizationOptimizationPass

OPTIMIZATION_LAYERTYPE_CONFIG = {
    1: {
        'Relu',
        'MaxPool',
        'GlobalMaxPool',
        'PRelu',
        'AveragePool',
        'GlobalAveragePool',
        'LeakyRelu',
        'Identity',
    },  # level - 1 optimize
    2: {
        'Relu',
        'MaxPool',
        'GlobalMaxPool',
        'Add',
        'Sub',
        'PRelu',
        'AveragePool',
        'GlobalAveragePool',
        'LeakyRelu',
        'Identity',
    },  # level - 2 optimize
}
EQUALIZATION_OPERATION_TYPE = {'Conv', 'Gemm', 'ConvTranspose'}


class ActivationEqualizationPass(QuantizationOptimizationPass):
    def __init__(
        self, interested_layers: List[str] = None, threshold: float = 4.0, optimize_level: int = 1, iterations: int = 10
    ) -> None:
        """PPQ Customized Layerwise Equalization Pass.

        This is a highly customized implementation of layerwise equalization.
        With PPQ graph searching engine, this implementation can equalize multiple layer at once,
            Even some layers are behind Add, Sub, Pooling.

        Not only weight, bias and activation are also taken into consideration with this implementation.
        if including_bias and including_activation set as True, all weight, bias, activation will be pull
            equal with this function.

        Args:
            iterations (int): Equalization iterations.
            weight_threshold (float, optional):
                Value threshold, all weight that belows that value will keep unchanged through this function.
                ATTENTION: this threshold will greatly affects your equalization performance.
                Defaults to 0.5. recommend to try 0.5, 2, 0

            including_bias (bool, optional):
                whether to include bias into consideration.
                ATTENTION: Some hardware use int16 accumulator or even worse
                    set this to be True if your hardware does not allow a 32-bit bias.
                Defaults to False.

            including_act (bool, optional):
                whether to include activation into consideration.
                Defaults to False.

            bias_multiplier (float, optional):
                a multiplier to bias, if not necessary do not change this.
                Defaults to 0.5.

            act_multiplier (float, optional):
                a multiplier to activation, if not necessary do not change this.
                Defaults to 0.5.

            interested_layers (List[str]):
                a layer collection contains all layers which need to be equalized.
                if None is given, suppose all layer need to be equalized.

            optimize_level (int, optional): [description]. Defaults to 2.
            verbose (bool, optional): [description]. Defaults to False.
        """
        self.interested_layers = interested_layers
        self.threshold = threshold
        self.optimize_level = optimize_level
        self.iterations = iterations
        super().__init__(name='PPQ Activation Equalization Pass')

    def find_equalization_pair(
        self, graph: BaseGraph, interested_operations: List[Operation]
    ) -> List[EqualizationPair]:
        # create a PPQ graph search engine.
        search_engine = SearchableGraph(graph)

        visited_ops, pairs = set(), []
        for operation in interested_operations:
            if operation in visited_ops:
                continue

            # skip operation that can not be equalized
            if operation.type not in EQUALIZATION_OPERATION_TYPE:
                continue

            # forward matching equalization pair.
            forward_matchings = search_engine(
                TraversalCommand(
                    sp_expr=lambda x: x == operation,
                    rp_expr=lambda x, y: y.type in OPTIMIZATION_LAYERTYPE_CONFIG[self.optimize_level],
                    ep_expr=lambda x: x.type not in OPTIMIZATION_LAYERTYPE_CONFIG[self.optimize_level],
                    direction='down',
                )
            )

            downstream_ops = set()
            for matching in forward_matchings:
                downstream_ops.add(matching[-1])

            # backward matching equalization pair
            forward_matchings = search_engine(
                TraversalCommand(
                    sp_expr=lambda x: x in downstream_ops,
                    rp_expr=lambda x, y: y.type in OPTIMIZATION_LAYERTYPE_CONFIG[self.optimize_level],
                    ep_expr=lambda x: x.type not in OPTIMIZATION_LAYERTYPE_CONFIG[self.optimize_level],
                    direction='up',
                )
            )

            upstream_ops = set()
            for matching in forward_matchings:
                upstream_ops.add(matching[-1])

            # update pairs to visited.
            visited_ops.update(upstream_ops)

            # check if all operation inside this pair can be properly processed.
            valid_flag = True
            for operation in upstream_ops:
                if operation.type not in EQUALIZATION_OPERATION_TYPE:
                    valid_flag = False

            for operation in downstream_ops:
                if operation.type not in EQUALIZATION_OPERATION_TYPE:
                    valid_flag = False

            if not valid_flag:
                continue

            # construct a new equalization pair.
            if len(upstream_ops) > 0 and len(downstream_ops) > 0:
                pairs.append(
                    EqualizationPair(upstream_layers=list(upstream_ops), downstream_layers=list(downstream_ops))
                )
        return pairs

    def collect_activations(
        self,
        graph: BaseGraph,
        executor: TorchExecutor,
        dataloader: Iterable,
        collate_fn: Callable,
        operations: List[Operation],
        steps: int = 16,
    ) -> Dict[str, torch.Tensor]:
        def aggregate(op: Operation, tensor: torch.Tensor):
            if op.type in {'Conv', 'ConvTranspose'}:  # Conv result: [n, c, h, w]
                num_of_channel = tensor.shape[1]
                tensor = tensor.transpose(0, 1)
                tensor = tensor.reshape(shape=[num_of_channel, -1])
                tensor = torch.max(tensor.abs(), dim=-1, keepdim=False)[0]
            elif op.type in {'MatMul', 'Gemm'}:  # Gemm result: [n, c]
                tensor = tensor.transpose(0, 1)
                tensor = torch.max(tensor.abs(), dim=-1, keepdim=False)[0]
            return tensor

        output_names = []
        for operation in operations:
            assert operation.num_of_output == 1, f'Num of output of layer {operation.name} is supposed to be 1'
            output_names.append(operation.outputs[0].name)

        output_collector = defaultdict(list)
        for idx, batch in tqdm(
            enumerate(dataloader), desc='Equalization Data Collecting.', total=min(len(dataloader), steps)
        ):
            data = batch
            if collate_fn is not None:
                data = collate_fn(batch)
            outputs = executor.forward(data, output_names=output_names)
            for name, output in zip(output_names, outputs):
                op = graph.variables[name].source_op
                output_collector[name].append(aggregate(op, output).unsqueeze(-1))
            if idx > steps:
                break

        result = {}
        for name, output in zip(output_names, outputs):
            result[name] = torch.cat(output_collector[name], dim=-1)
        return result

    @empty_ppq_cache
    def optimize(
        self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, collate_fn: Callable, **kwargs
    ) -> None:
        interested_operations = []

        if self.interested_layers is None:
            for operation in graph.operations.values():
                if operation.type in EQUALIZATION_OPERATION_TYPE:
                    interested_operations.append(operation)
        else:
            for name in self.interested_layers:
                if name in graph.operations:
                    interested_operations.append(graph.operations[name])

        pairs = self.find_equalization_pair(graph=graph, interested_operations=interested_operations)

        activations = self.collect_activations(
            graph=graph,
            executor=executor,
            dataloader=dataloader,
            collate_fn=collate_fn,
            operations=interested_operations,
        )

        for name, act in activations.items():
            graph.variables[name].value = act  # 将激活值写回网络

        print(f'{len(pairs)} equalization pair(s) was found, ready to run optimization.')
        for iter_times in tqdm(range(self.iterations), desc='Activation Equalization', total=self.iterations):
            for equalization_pair in pairs:
                equalization_pair.activation_equalize(threshold=self.threshold)

        # equalization progress directly changes fp32 value of weight,
        # store it for following procedure.
        for op in graph.operations.values():
            if isinstance(op, QuantableOperation):
                op.store_parameter_value()


class LayerwiseEqualizationPass(QuantizationOptimizationPass):
    """
    ## Layer-wise Equalization Pass(层间权重均衡过程)

    Weight distributions can differ strongly between output channels,
    using only one quantization scale, per-tensor quantization has its trouble for representing the value among channels.

    For example, in the case where one channel has weights in the range [−128, 128] and another channel has weights in the range (−0.5, 0.5),
    the weights in the latter channel will all be quantized to 0 when quantizing to 8-bits.

    Hopefully, the performance can be improved by adjusting the weights for each output channel such that their ranges are more similar.

    Formula:

            Take 2 convolution layers as an example

            Where Y = W_2 * (W_1 * X + b_1) + b_2

            Adjusting W_1, W_2 by a scale factor s:

            Y = W_2 / s * (W_1 * s * X + b_1 * s) + b_2

            Where s has the same dimension as the output channel of W_1

    This method is called as Layer-wise Equalization, which is proposed by Markus Nagel.

    https://openaccess.thecvf.com/content_ICCV_2019/papers/Nagel_Data-Free_Quantization_Through_Weight_Equalization_and_Bias_Correction_ICCV_2019_paper.pdf

    self, iterations: int, weight_threshold: float = 0.5,
    including_bias: bool = False, including_activation: bool = False,
    bias_multiplier: float = 0.5, activation_mutiplier: float = 0.5,
    interested_layers: List[str] = None, optimize_level: int = 2,
    verbose:bool = False

    ### Parameters:

    * iterations(int):

            Integer value of Algorithm iterations.

            More iterations will give more plainness in your weight distribution,
            iteration like 100 can flatten all the parameter in your network to a same level.

            You are not recommended to iterate until value converges,
            in some cases stop iteration earlier will give you a better performance.

    * weight_threshold(float)

            A threshold that stops processing value that is too small.

            By default, the scale factor of equalization method is computed as sqrt(max(abs(W_1)) / max(abs(W_2))),
            the maximum value of W_2 can be very small(like 1e-14), while the maximum value W_1 can be 0.5.

            In this case, the computed scale factor is 1e7, the optimization will loss its numerical stability and even give an unreasonable result.

            To prevent the scale factor becoming too large, ppq clips all the value smaller than this threshold before iterations.

            This parameter will significantly affects the optimization result.

            Recommended values are 0, 0.5, 2.

    # including_bias(bool)

            Whether to include bias in computing scale factor.

            If including_bias is True, the scale factor will be computed as sqrt(max(abs(W_1 : b_1)) / max(abs(W_2 : b_2)))

            Where W_1 : b_1 mean an augmented matrix with W_1 and b_1

    # including_bias(float)

            Only take effects when including_bias = True

            the scale factor will be computed as sqrt(max(abs(W_1 : b_1 * bias_multiplier)) / max(abs(W_2 : b_2 * bias_multiplier)))

            This is an correction term for bias.

    # including_activation(bool)

            Same as the parameter including_bias, whether to include activation in computing scale factor.

    # activation_multiplier(float)

            Same as the including_bias, this is an correction term for activation.

    # optimize_level(int)

            level - 1: equalization will only cross ('Relu', 'MaxPool', 'GlobalMaxPool', 'PRelu', 'AveragePool', 'GlobalAveragePool')

            level - 2: equalization will cross ('Relu', 'MaxPool', 'GlobalMaxPool', 'Add', 'Sub', 'PRelu', 'AveragePool', 'GlobalAveragePool')

            Here is an example for illustrating the difference, if we got a graph like:

                Conv1 - Relu - Conv2

            Both level - 1 and level - 2 optimization can find there is a equalization pair: (Conv1 - Conv2).

            however for a complex graph like:

                Conv1 - Add - Conv2

            level - 1 optimization will simply skip Conv1 and Conv2.

            level - 2 optimization will trace another input from Add, and then PPQ will take all the input operations of Add
            as the upstream layers in equalization.

            PPQ use graph search engine for pasring graph structure, check esp_ppq.IR.search.py for more information.

    # interested_layers(List[str])

            Only layer that listed in interested_layers will be processed by this pass.

            If interested_layers is None or empty list, all the layers will be processed.

    ### Warning:
    You can not compare a equalized graph with an unequalized graph layer by layer,
    since equalization pass guarantees only the output of your network will be kept as same,
    the intermediate result can be changed rapidly.

    Since then, PPQ invokes this pass before network quantization.

    ### Usage
    Layer-wise equalization are designed for per-layer quantization.

    |              | Symmetrical | Asymmetrical | Per-chanel    | Per-tensor |
    | ---          | ---         | ---          | ---           | ---        |
    | equalizaiton |             |              | Not recommend |            |

    Layer-wise Equalization Pass should be invoked in pre-quant optimization pipeline.
    (pre-quant optimization pipeline take effects before quantization)

    This pass is included in PPQ Quantization Setting, you can calling this optimization by:

        setting = QuantizationSettingFactory.default_setting()

        setting.equalization = True

        # calling esp_ppq.api.quantize_onnx_model function with this setting.
        ir = quantize_torch_model(
        model=model, calib_dataloader=load_calibration_dataset(), setting=setting,
        platform=TargetPlatform.PPL_CUDA_INT8, calib_steps=8, input_shape=INPUT_SHAPE,
        collate_fn=collate_fn)

    You can manually create this optimization by:

        from esp_ppq import LayerwiseEqualizationPass

        optim = LayerwiseEqualizationPass()

    """

    def __init__(
        self,
        iterations: int,
        value_threshold: float = 0.5,
        including_weight: bool = True,
        weight_multiplier: float = 1.0,
        including_bias: bool = False,
        including_act: bool = False,
        bias_multiplier: float = 0.5,
        act_multiplier: float = 0.5,
        interested_layers: List[str] = None,
        optimize_level: int = 2,
        verbose: bool = False,
    ) -> None:
        """PPQ Customized Layerwise Equalization Pass.

        This is a highly customized implementation of layerwise equalization.
        With PPQ graph searching engine, this implementation can equalize multiple layer at once,
            Even some layers are behind Add, Sub, Pooling.

        Not only weight, bias and activation are also taken into consideration with this implementation.
        if including_bias and including_activation set as True, all weight, bias, activation will be pull
            equal with this function.

        Args:
            iterations (int): Equalization iterations.
            weight_threshold (float, optional):
                Value threshold, all weight that belows that value will keep unchanged through this function.
                ATTENTION: this threshold will greatly affects your equalization performance.
                Defaults to 0.5. recommend to try 0.5, 2, 0

            including_bias (bool, optional):
                whether to include bias into consideration.
                ATTENTION: Some hardware use int16 accumulator or even worse
                    set this to be True if your hardware does not allow a 32-bit bias.
                Defaults to False.

            including_act (bool, optional):
                whether to include activation into consideration.
                Defaults to False.

            bias_multiplier (float, optional):
                a multiplier to bias, if not necessary do not change this.
                Defaults to 0.5.

            act_multiplier (float, optional):
                a multiplier to activation, if not necessary do not change this.
                Defaults to 0.5.

            interested_layers (List[str]):
                a layer collection contains all layers which need to be equalized.
                if None is given, suppose all layer need to be equalized.

            optimize_level (int, optional): [description]. Defaults to 2.
            verbose (bool, optional): [description]. Defaults to False.
        """
        self.optimize_level = optimize_level
        self.iterations = iterations
        self.value_threshold = value_threshold

        self.including_weight = including_weight
        self.weight_multiplier = weight_multiplier

        self.including_bias = including_bias
        self.bias_multiplier = bias_multiplier

        self.including_act = including_act
        self.act_multiplier = act_multiplier

        self.interested_layers = interested_layers
        self.verbose = verbose
        super().__init__(name='PPQ Layerwise Equalization Pass')

    def find_equalization_pair(
        self, graph: BaseGraph, interested_operations: List[Operation]
    ) -> List[EqualizationPair]:
        # create a PPQ graph search engine.
        search_engine = SearchableGraph(graph)

        visited_ops, pairs = set(), []
        for operation in interested_operations:
            if operation in visited_ops:
                continue

            # skip operation that can not be equalized
            if operation.type not in EQUALIZATION_OPERATION_TYPE:
                continue

            # forward matching equalization pair.
            forward_matchings = search_engine(
                TraversalCommand(
                    sp_expr=lambda x: x == operation,
                    rp_expr=lambda x, y: y.type in OPTIMIZATION_LAYERTYPE_CONFIG[self.optimize_level],
                    ep_expr=lambda x: x.type not in OPTIMIZATION_LAYERTYPE_CONFIG[self.optimize_level],
                    direction='down',
                )
            )

            downstream_ops = set()
            for matching in forward_matchings:
                downstream_ops.add(matching[-1])

            # backward matching equalization pair
            forward_matchings = search_engine(
                TraversalCommand(
                    sp_expr=lambda x: x in downstream_ops,
                    rp_expr=lambda x, y: y.type in OPTIMIZATION_LAYERTYPE_CONFIG[self.optimize_level],
                    ep_expr=lambda x: x.type not in OPTIMIZATION_LAYERTYPE_CONFIG[self.optimize_level],
                    direction='up',
                )
            )

            upstream_ops = set()
            for matching in forward_matchings:
                upstream_ops.add(matching[-1])

            # update pairs to visited.
            visited_ops.update(upstream_ops)

            # check if all operation inside this pair can be properly processed.
            valid_flag = True
            for operation in upstream_ops:
                if operation.type not in EQUALIZATION_OPERATION_TYPE:
                    valid_flag = False

            for operation in downstream_ops:
                if operation.type not in EQUALIZATION_OPERATION_TYPE:
                    valid_flag = False

            if not valid_flag:
                continue

            # construct a new equalization pair.
            if len(upstream_ops) > 0 and len(downstream_ops) > 0:
                pairs.append(
                    EqualizationPair(upstream_layers=list(upstream_ops), downstream_layers=list(downstream_ops))
                )
        return pairs

    def collect_activations(
        self,
        graph: BaseGraph,
        executor: TorchExecutor,
        dataloader: Iterable,
        collate_fn: Callable,
        operations: List[Operation],
        steps: int = 16,
    ) -> Dict[str, torch.Tensor]:
        def aggregate(op: Operation, tensor: torch.Tensor):
            if op.type in {'Conv', 'ConvTranspose'}:  # Conv result: [n, c, h, w]
                num_of_channel = tensor.shape[1]
                tensor = tensor.transpose(0, 1)
                tensor = tensor.reshape(shape=[num_of_channel, -1])
                tensor = torch.max(tensor.abs(), dim=-1, keepdim=False)[0]
            elif op.type in {'MatMul', 'Gemm'}:  # Gemm result: [n, c]
                tensor = tensor.transpose(0, 1)
                tensor = torch.max(tensor.abs(), dim=-1, keepdim=False)[0]
            return tensor

        output_names = []
        for operation in operations:
            assert operation.num_of_output == 1, f'Num of output of layer {operation.name} is supposed to be 1'
            output_names.append(operation.outputs[0].name)

        output_collector = defaultdict(list)
        for idx, batch in tqdm(
            enumerate(dataloader), desc='Equalization Data Collecting.', total=min(len(dataloader), steps)
        ):
            data = batch
            if collate_fn is not None:
                data = collate_fn(batch)
            outputs = executor.forward(data, output_names=output_names)
            for name, output in zip(output_names, outputs):
                op = graph.variables[name].source_op
                output_collector[name].append(aggregate(op, output).unsqueeze(-1))
            if idx > steps:
                break

        result = {}
        for name, output in zip(output_names, outputs):
            result[name] = torch.cat(output_collector[name], dim=-1)
        return result

    @empty_ppq_cache
    def optimize(
        self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, collate_fn: Callable, **kwargs
    ) -> None:
        interested_operations = []

        if self.interested_layers is None or len(self.interested_layers) == 0:
            for operation in graph.operations.values():
                if operation.type in EQUALIZATION_OPERATION_TYPE:
                    interested_operations.append(operation)
        else:
            for name in self.interested_layers:
                if name in graph.operations:
                    interested_operations.append(graph.operations[name])

        pairs = self.find_equalization_pair(graph=graph, interested_operations=interested_operations)

        if self.including_act:
            activations = self.collect_activations(
                graph=graph,
                executor=executor,
                dataloader=dataloader,
                collate_fn=collate_fn,
                operations=interested_operations,
            )

            for name, act in activations.items():
                graph.variables[name].value = act  # 将激活值写回网络

        print(f'{len(pairs)} equalization pair(s) was found, ready to run optimization.')
        for iter_times in tqdm(range(self.iterations), desc='Layerwise Equalization', total=self.iterations):
            for equalization_pair in pairs:
                print([layer for layer in equalization_pair.upstream_layers])
                equalization_pair.equalize(
                    value_threshold=self.value_threshold,
                    including_weight=self.including_weight,
                    weight_multiplier=self.weight_multiplier,
                    including_bias=self.including_bias,
                    including_act=self.including_act,
                    bias_multiplier=self.bias_multiplier,
                    act_multiplier=self.act_multiplier,
                )

        # equalization progress directly changes fp32 value of weight,
        # store it for following procedure.
        for op in graph.operations.values():
            if isinstance(op, QuantableOperation):
                op.store_parameter_value()


class ChannelwiseSplitPass(LayerwiseEqualizationPass):
    def __init__(
        self,
        iterations: int,
        threshold: float = 2,
        including_bias: bool = False,
        bias_multiplier: float = 0.5,
        including_act: bool = False,
        act_multiplier: float = 0.5,
        interested_layers: List[str] = None,
        optimize_level: int = 2,
        verbose: bool = False,
    ) -> None:
        self.optimize_level = optimize_level
        self.iterations = iterations
        self.value_threshold = threshold

        self.including_bias = including_bias
        self.bias_multiplier = bias_multiplier

        self.including_act = including_act
        self.act_multiplier = act_multiplier

        self.interested_layers = interested_layers
        self.verbose = verbose

        self.name = 'PPQ Channelwise Split Pass'

    def optimize(
        self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, collate_fn: Callable, **kwargs
    ) -> None:
        interested_operations = []
        if self.interested_layers is None or len(self.interested_layers) == 0:
            for operation in graph.operations.values():
                if operation.type in EQUALIZATION_OPERATION_TYPE:
                    interested_operations.append(operation)
        else:
            for name in self.interested_layers:
                if name in graph.operations:
                    interested_operations.append(graph.operations[name])

        pairs = self.find_equalization_pair(graph=graph, interested_operations=interested_operations)

        print(f'{len(pairs)} equalization pair(s) was found, ready to run optimization.')
        for iter_times in tqdm(range(self.iterations), desc='Layerwise Channel Split', total=self.iterations):
            if self.including_act:
                activations = self.collect_activations(
                    graph=graph,
                    executor=executor,
                    dataloader=dataloader,
                    collate_fn=collate_fn,
                    operations=interested_operations,
                )

                for name, act in activations.items():
                    graph.variables[name].value = act  # 将激活值写回网络

            for equalization_pair in pairs:
                # can not split group convolution.
                is_group_conv = False
                for layer in equalization_pair.downstream_layers + equalization_pair.upstream_layers:
                    if layer.type in {'Conv', 'ConvTranspose'}:
                        group = layer.attributes.get('group', 1)
                        if group != 1:
                            is_group_conv = True

                if is_group_conv:
                    continue
                equalization_pair.channel_split(
                    value_threshold=self.value_threshold,
                    including_bias=self.including_bias,
                    including_act=self.including_act,
                    bias_multiplier=self.bias_multiplier,
                    act_multiplier=self.act_multiplier,
                )

        # channel split progress directly changes fp32 value of weight,
        # store it for following procedure.
        for op in graph.operations.values():
            if isinstance(op, QuantableOperation):
                op.store_parameter_value()



# ---------------------------------------------------------------
# 这个脚本向你展示了如何使用 tensorRT 对 PPQ 导出的模型进行推理，并进行速度测试
# 目前 GPU 上 tensorRT 是跑的最快的部署框架 ...
# ---------------------------------------------------------------

import time

import numpy as np
import tensorrt as trt
import torch
import torchvision
import torchvision.models
import trt_infer
from tqdm import tqdm

from esp_ppq import *
from esp_ppq.api import *

# Nvidia Nsight Performance Profile
QUANT_PLATFROM = TargetPlatform.TRT_INT8
BATCHSIZE = 1
SAMPLES = [torch.zeros(size=[BATCHSIZE, 3, 640, 640]) for _ in range(32)]
DEVICE = 'cuda'
MODEL_PATH = 'models/yolov6s.onnx'
CFG_VALID_RESULT = False


def infer_trt(model_path: str, samples: List[np.ndarray]) -> List[np.ndarray]:
    """Run a tensorrt model with given samples"""
    logger = trt.Logger(trt.Logger.ERROR)
    with open(model_path, 'rb') as f, trt.Runtime(logger) as runtime:
        engine = runtime.deserialize_cuda_engine(f.read())

    results = []
    if CFG_VALID_RESULT:
        with engine.create_execution_context() as context:
            inputs, outputs, bindings, stream = trt_infer.allocate_buffers(context.engine)
            for sample in tqdm(samples, desc='TensorRT is running...'):
                inputs[0].host = convert_any_to_numpy(sample)
                [output] = trt_infer.do_inference(
                    context, bindings=bindings, inputs=inputs, outputs=outputs, stream=stream, batch_size=1
                )[0]
                results.append(convert_any_to_torch_tensor(output).reshape([-1, 1000]))
    else:
        with engine.create_execution_context() as context:
            inputs, outputs, bindings, stream = trt_infer.allocate_buffers(context.engine)
            inputs[0].host = convert_any_to_numpy(samples[0])
            for sample in tqdm(samples, desc='TensorRT is running...'):
                trt_infer.do_inference(
                    context, bindings=bindings, inputs=inputs, outputs=outputs, stream=stream, batch_size=1
                )
    return results


with ENABLE_CUDA_KERNEL():
    # export non-quantized model to tensorRT for benchmark
    non_quantized = quantize_onnx_model(
        onnx_import_file=MODEL_PATH,
        calib_dataloader=SAMPLES,
        collate_fn=lambda x: x.to(DEVICE),
        calib_steps=32,
        input_shape=[BATCHSIZE, 3, 640, 640],
        setting=QuantizationSettingFactory.default_setting(),
        platform=QUANT_PLATFROM,
        do_quantize=False,
    )

    export_ppq_graph(graph=non_quantized, platform=TargetPlatform.ONNX, graph_save_to='model_fp32.onnx')
    builder = trt_infer.EngineBuilder()
    builder.create_network('model_fp32.onnx')
    builder.create_engine(engine_path='model_fp32.engine', precision="fp16")

    # quantize model with esp_ppq.
    quantized = quantize_onnx_model(
        onnx_import_file=MODEL_PATH,
        calib_dataloader=SAMPLES,
        collate_fn=lambda x: x.to(DEVICE),
        calib_steps=32,
        input_shape=[BATCHSIZE, 3, 640, 640],
        setting=QuantizationSettingFactory.default_setting(),
        platform=QUANT_PLATFROM,
    )

    if CFG_VALID_RESULT:
        executor = TorchExecutor(graph=quantized)
        ref_results = []
        for sample in tqdm(SAMPLES, desc='PPQ GENERATEING REFERENCES', total=len(SAMPLES)):
            result = executor.forward(inputs=sample.to(DEVICE))[0]
            result = result.cpu().reshape([-1, 1000])
            ref_results.append(result)

    # export model to disk.
    export_ppq_graph(graph=quantized, platform=TargetPlatform.TRT_INT8, graph_save_to='model_int8.onnx')

    if CFG_VALID_RESULT:
        # compute simulating error
        trt_outputs = infer_trt(
            model_path='model_int8.engine', samples=[convert_any_to_numpy(sample) for sample in SAMPLES]
        )

        error = []
        for ref, real in zip(ref_results, trt_outputs):
            ref = convert_any_to_torch_tensor(ref).float()
            real = convert_any_to_torch_tensor(real).float()
            error.append(torch_snr_error(ref, real))
        error = sum(error) / len(error) * 100
        print(f'Simulating Error: {error: .4f}%')

    # benchmark with onnxruntime int8
    benchmark_samples = [np.zeros(shape=[BATCHSIZE, 3, 224, 224], dtype=np.float32) for _ in range(512)]
    print(f'Start Benchmark with tensorRT (Batchsize = {BATCHSIZE})')
    tick = time.time()
    infer_trt(model_path='model_fp32.engine', samples=benchmark_samples)
    tok = time.time()
    print(f'Time span (FP32 MODE): {tok - tick: .4f} sec')

    tick = time.time()
    infer_trt(model_path='model_int8.engine', samples=benchmark_samples)
    tok = time.time()
    print(f'Time span (INT8 MODE): {tok - tick: .4f} sec')



from abc import ABCMeta, abstractmethod
from typing import Any

from esp_ppq.core import QuantizationStates, TensorQuantizationConfig, ppq_debug_function
from esp_ppq.IR import Variable


class BaseTensorObserver(metaclass=ABCMeta):
    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        self._watch_on = watch_on
        self._quant_cfg = quant_cfg

    @abstractmethod
    def observe(self, value: Any):
        raise NotImplementedError('Implement this function first.')

    @abstractmethod
    def render_quantization_config(self):
        raise NotImplementedError('Implement this function first.')

    def __str__(self) -> str:
        return (
            'PPQ Tensor Observer ('
            + self.__class__.__name__
            + ') mount on variable '
            + self._watch_on.name
            + ' observing algorithm: '
            + self._quant_cfg.observer_algorithm
        )

    @ppq_debug_function
    def report(self) -> str:
        if self._quant_cfg.state == QuantizationStates.ACTIVATED:
            return (
                f'Observer on Variable {self._watch_on.name}, '
                f'computed scale: {self._quant_cfg.scale}, computed offset: {self._quant_cfg.offset}\n'
            )



"""PPQ Core Decorator & MetaClass definitions PPQ 核心装饰器、元类型定义.

You are not allowed to modify this 请勿修改此文件
"""

import gc
from typing import Callable

from torch.cuda import empty_cache

from esp_ppq.log import NaiveLogger

from .config import PPQ_CONFIG

logger = NaiveLogger.get_logger('PPQ')


class SingletonMeta(type):
    """The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.

    see also: https://refactoring.guru/design-patterns/singleton/python/example
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Possible changes to the value of the `__init__` argument do not
        affect the returned instance."""
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


def ppq_legacy(func: str, version: str, adapt_to: str = None):
    """Mark an function as legacy function.

    Args:
        func (str): _description_
        version (str): _description_
        adapt_to (str, optional): _description_. Defaults to None.
    """
    print(f'{func} has been obsoleted since PPQ {version}, use {adapt_to} instead.')


def empty_ppq_cache(func: Callable):
    """Using empty_ppq_cache decorator to clear ppq memory cache, both gpu
    memory and cpu memory will be clear via this function.

    Function which get decorated by this will clear all ppq system cache BEFORE its running.
    Args:
        func (Callable): decorated function
    """

    def _wrapper(*args, **kwargs):
        empty_cache()  # torch.cuda.empty_cache might requires a sync of all cuda device.
        gc.collect()  # empty memory.
        return func(*args, **kwargs)

    return _wrapper


def ppq_quant_param_computing_function(func: Callable):
    """mark a function to be a scale-computing function.

    Args:
        func (Callable): decorated function
    """

    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return _wrapper


def ppq_debug_function(func: Callable):
    """mark a function to be a debug function.

    Args:
        func (Callable): decorated function
    """

    def _wrapper(*args, **kwargs):
        if PPQ_CONFIG.PPQ_DEBUG:
            debug_str = func(*args, **kwargs)
            if debug_str is None:
                return None
            assert isinstance(debug_str, str), (
                f'ppq_debug_function should only return string instance, while {str(func)} returns {type(debug_str)}'
            )
            print(debug_str, end='')
        else:
            return None

    return _wrapper


def ppq_file_io(func: Callable):
    """mark a function to be a ppq file io function.

    function must have return a file handle.
    Args:
        func (Callable): decorated function
    """

    def _wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return _wrapper


def ppq_warning(info: str):
    logger.warning(info)


def ppq_info(info: str):
    logger.info(info)



# ---------------------------------------------------------------
# 这个脚本向你展示了如何使用 torch2trt 加速 pytorch 推理
# 截止目前为止 torch2trt 的适配能力有限，不要尝试运行特别奇怪的模型
# 你可以把模型分块来绕开那些不支持的算子。

# 使用之前你必须先装好 TensorRT, torch2trt等工具包
# https://github.com/NVIDIA-AI-IOT/torch2trt

# ---------------------------------------------------------------

import torch
import torch.utils.data
import torchvision
from torch2trt import torch2trt
from tqdm import tqdm

SAMPLES = [torch.zeros(1, 3, 224, 224) for _ in range(1024)]
MODEL = torchvision.models.resnet18()
FP16_MODE = True

# Model has to be the eval mode, and deploy to cuda.
MODEL.eval()
MODEL.cuda()


def trace_handler(prof):
    print(prof.key_averages().table(sort_by="self_cuda_time_total", row_limit=-1))


# Benckmark with pytorch
for sample in tqdm(SAMPLES, desc='Torch Executing'):
    MODEL.forward(sample.cuda())

# Convert torch.nn.Module to tensorrt
# 在转换过后，你模型中的执行函数将会被 trt 替换，同时进行图融合
model_trt = torch2trt(MODEL, [sample.cuda()], fp16_mode=FP16_MODE)
for sample in tqdm(SAMPLES, desc='TRT Executing'):
    model_trt.forward(sample.cuda())

# Test performance metrics using torch.profiler
with torch.profiler.profile(
    activities=[torch.profiler.ProfilerActivity.CPU, torch.profiler.ProfilerActivity.CUDA],
    schedule=torch.profiler.schedule(wait=2, warmup=1, active=7),
    on_trace_ready=trace_handler,
    # on_trace_ready=torch.profiler.tensorboard_trace_handler('log')
    # used when outputting for tensorboard
) as p:
    for iter in range(10):
        model_trt.forward(sample.cuda())
        # send a signal to the profiler that the next iteration has started
        p.step()

with torch.profiler.profile(
    activities=[torch.profiler.ProfilerActivity.CPU, torch.profiler.ProfilerActivity.CUDA],
    schedule=torch.profiler.schedule(wait=2, warmup=1, active=7),
    on_trace_ready=trace_handler,
    # on_trace_ready=torch.profiler.tensorboard_trace_handler('log')
    # used when outputting for tensorboard
) as p:
    for iter in range(10):
        MODEL.forward(sample.cuda())
        # send a signal to the profiler that the next iteration has started
        p.step()



from math import pow


class EMARecorder:
    """Exponential Moving Average(EMA) with bias correction."""

    def __init__(self, beta: float = 0.98):
        self.beta = beta
        self.t = 0
        self.value = 0

    def push(self, value: float):
        self.value = (1.0 - self.beta) * value + self.beta * self.value
        self.t += 1

    def pop(self) -> float:
        if self.t == 0:
            return 0
        return self.value / (1 - pow(self.beta, self.t))



from typing import Callable, Dict, List, Union

import torch

from esp_ppq.core import (
    DataType,
    QuantizationStates,
    TargetPlatform,
    TensorQuantizationConfig,
    empty_ppq_cache,
    ppq_warning,
)
from esp_ppq.IR import BaseGraph, Operation, QuantableOperation, RunnableGraph
from esp_ppq.IR.base.command import GraphDeployCommand
from esp_ppq.quantization.qfunction import PPQuantFunction

from .base import OPERATION_FORWARD_TABLE, BaseGraphExecutor, QuantOPRuntimeHook, RuntimeHook
from .op import TorchBackendContext


class TorchMetaDataTracingHook(RuntimeHook):
    def __init__(self, operation: Operation) -> None:
        super().__init__(operation)

    def pre_forward_hook(self, inputs: List[torch.Tensor], **kwargs) -> list:
        # some operations got none as its input
        # therefore we have to create meta for those none input value manually.
        for tensor, var in zip(inputs, self._hook_to.inputs):
            if tensor is None:
                ppq_warning(
                    f'Unexpected input value of operation {self._hook_to.name}, '
                    f'recieving "None" at its input {self._hook_to.inputs.index(var)}'
                )
            else:
                var.shape = tensor.shape
                var.dtype = tensor.dtype

        return inputs

    def post_forward_hook(self, outputs: List[torch.Tensor], **kwargs) -> list:
        for tensor, var in zip(outputs, self._hook_to.outputs):
            if tensor is not None:
                var.shape = tensor.shape
                var.dtype = tensor.dtype

        return outputs


class TorchQuantizeDelegator(Callable):
    """Since PPQ 0.6.2, Interface TorchQuantizeDelegate is introduced to
    customize quantization logic: To be specific, you are suppose to inherit
    this class, and define your own computation logic within function __call__.

        Pass your Delegate to TorchExecutor by TorchExecutor.register_quantize_delegate(c, d)
            Where c is the target quantization config, d is your delegator class.
            Once you invoke this function, PPQ execution system will hand the quantization
            computation of config c over to your delegate. PPQ execution system will no
            longer quantize variable related with config c anymore.

    Notice that a delegate replaces quantization computation only, it still under the control of PPQ quantization
    System, so to say if your config has an invalid state like DEQUANTIZED, PPQ execution system will never been
    required to quantize related tensor and so your delegate class will take no effects on config c.

        Remove delegate function by TorchExecutor.remove_quantize_delegate(c)

    If you have some customized parameter of your delegator logic, set them as class attributes.
    Like: self.param1 = ..., self.param2 = ...

    Do not edit config structure directly.

    Args:
        Callable (_type_): _description_
    """

    def __init__(self) -> None:
        super().__init__()

    def __call__(self, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
        raise NotImplementedError('Implement this function first.')


class TorchExecutor(BaseGraphExecutor, torch.nn.Module):
    def __init__(self, graph: BaseGraph, fp16_mode: bool = True, device: str = 'cuda') -> None:
        """
        ## PPQ Graph Executor(PPQ 执行引擎)

        为了量化并优化神经网络模型，PPQ 实现了基于 Pytorch 的执行引擎，该执行引擎能够执行 Onnx 的模型文件，目前支持 90 余种常见 Onnx 算子，涵盖 1d, 2d, 3d 视觉、语音、文本模型。
        PPQ 的执行引擎位于 esp_ppq.executor 目录下，由两个主要部分组成： esp_ppq.executor.torch.py 文件中包含了执行引擎自身； esp_ppq.executor.op 文件夹中则包含了不同后端的算子库。

        在开始阅理解执行引擎之前，我们先介绍算子库的相关内容

        ### PPQ Backend Functions(PPQ 算子库)

        核心算子库位于 esp_ppq.executor.op.torch.default 文件中，该文件中包含了所有算子默认的执行逻辑。

        我们知道，对于一个量化算子而言，由于硬件的不同其执行逻辑也可能发生变化。例如 LeakyRelu 算子的负数部分在 GPU 上会采用 x * alpha 的方式完成计算，
        而在 FPGA 则会采用 x = x >> 3 完成计算。正因为这种差异的存在， PPQ 允许相同的算子在不同平台(TargetPlatform)上拥有不同的执行逻辑。
        这也意味着针对每一个平台，我们都将实现一个平台独特的算子库文件，这些算子库都继承于 esp_ppq.executor.op.torch.default。

            def Mul_forward(op: Operation, values: List[torch.Tensor], ctx: TorchBackendContext = None, **kwargs) -> torch.Tensor:
                ASSERT_NUM_OF_INPUT(op=op, values=values, min_num_of_input=2, max_num_of_input=2)
                values = VALUE_TO_EXECUTING_DEVICE(op=op, ctx=ctx, values=values)
                multiplicand, multiplier = values
                return multiplicand * multiplier

        上文中的内容即 esp_ppq.executor.op.torch.default 中 Mul 算子的执行逻辑，在 PPQ 中，所有算子在执行时都将接受一系列 torch.Tensor 作为输入，而后我们调用 pytorch 完成算子的计算逻辑。
        你可以打开 PPQ 的算子库文件查看其他算子的执行逻辑，并且 PPQ 也提供了 register_operation_handler 函数，借助该函数你可以注册自定义算子的执行逻辑；或是覆盖现有算子的执行逻辑。

            def register_operation_handler(handler: Callable, operation_type: str, platform: TargetPlatform):
                if platform not in GLOBAL_DISPATCHING_TABLE:
                    raise ValueError('Unknown Platform detected, Please check your platform setting.')
                GLOBAL_DISPATCHING_TABLE[platform][operation_type] = handler

        该函数位于 esp_ppq.api, 你可以使用语句 from esp_ppq.api import register_operation_handler 来引入它。

        ### PPQ Executor(PPQ 执行引擎)
        接下来我们向你介绍 PPQ 执行引擎 TorchExecutor，你可以使用语句 from ppq import TorchExecutor 导入执行引擎。初始化执行引擎则需要传入一个 PPQ 计算图实例对象，
        在这里我们假设已经获取到了一个量化后的计算图对象 ppq_quant_ir，并使用下面的语句初始化计算引擎


            executor = TorchExecutor(graph=ppq_quant_ir)
            executor.forward(inputs=..., output_names=..., hooks=...)

        我们使用 executor.forward 接口获取图的执行结果，它将可以传入三个参数:

        * inputs: inputs (Union[dict, list, torch.Tensor]): [input tensors or somewhat]
        * output_names (List[str], optional): output variable names. default is None.
        * hooks (Dict[str, RuntimeHook], optional): A hook table for customizing operation behaviour and collate data during executing.

        当执行引擎获取到推理请求时，它将按拓扑顺序依次执行图中的算子：下图展示了一个简单的示例

        ![图1. 执行图示例](https://user-images.githubusercontent.com/43309460/190056256-8b664993-3af4-4151-8451-f60d42f3145c.png)

        在这里，我们的图中包含三个算子 Conv, Relu, Softmax，他们将按照拓扑次序被依次执行。PPQ 的执行引擎会在执行完 Conv 算子后，将 Conv 算子的结果暂存于 Var 1 中，供 Relu 算子取用。
        而在执行完 Relu 算子后，PPQ 执行引擎则会及时地释放 Var 1 中暂存的数据，因为他们不会被其他算子取用，而且也不是网络的输出 Variable。在每一次推理过后，PPQ 还会清空网络中所有的暂存变量以释放显存。
        下面的代码段展示了一个非量化算子的执行逻辑：

            for operation in executing_order:
                outputs = operation_forward_func(operation, inputs, self._executing_context)
                outputs = outputs if isinstance(outputs, (list, tuple)) else [outputs]
                fp_outputs = outputs

                for output_idx, output_var in enumerate(operation.outputs):
                    output_var       = operation.outputs[output_idx]
                    output_var.value = outputs[output_idx]

            for var in self._graph.variables.values():
                if not var.is_parameter:
                    var.value = None

        PPQ 的执行引擎是专为量化计算图的执行而设计的————接下来让我们深入到量化算子的执行过程中去。
        对于一个量化算子而言，其每一个输入和输出变量都会有一个 Tensor Quantization Config (TQC) 控制结构体对量化过程进行描述。

        对于一个量化 Conv 算子而言，PPQ 将为它创建 2-3 个 Input TQC，以及一个 Output TQC。分别对其输入变量以及输出变量的量化行为进行描述。
        下面的代码展示了如何为量化算子创建特定的 TQC 描述量化逻辑。

            if operation.type == 'Conv':
                config = self.create_default_quant_config(
                    op                 = operation,
                    num_of_bits        = 8,
                    quant_max          = 127,
                    quant_min          = -128,
                    observer_algorithm = 'percentile',
                    policy             = QuantizationPolicy(
                        QuantizationProperty.PER_TENSOR +
                        QuantizationProperty.LINEAR +
                        QuantizationProperty.SYMMETRICAL),
                    rounding           = RoundingPolicy.ROUND_HALF_EVEN)

                for tensor_quant_config in config.input_quantization_config:
                    tensor_quant_config.state = QuantizationStates.FP32

                operation.config = config

        在 图2 中，我们展示了 PPQ 执行引擎对于量化算子的执行逻辑：

        ![图2. 算子执行流程](https://user-images.githubusercontent.com/43309460/190065996-02bf7fb4-7421-4417-883d-77967dd1863a.png)

        在 PPQ 中，算子的执行被分为四个过程：
        * 首先 PPQ 将根据算子上的 TQC 信息量化算子的输入。量化过程并非是原地的，量化后的数据将会是一个新的 torch.Tensor。
        * 随后 PPQ 在算子库中寻找算子的执行逻辑，我们已经提到对于每一个平台，他们都可以拥有自己的一套独立的算子库。
        PPQ 将按照算子的平台找到特定的计算逻辑，并调用他们完成计算得到结果。
        * PPQ 将根据算子上的 TQC 信息量化算子的输出。同样地，输出的量化也不是原地的。
        * 最后我们将量化好的结果写入到计算图的 Variable 上，从而供后续的算子取用。

        对于一个非量化算子而言，上述步骤中的 1，3 是可以省略的。

        下 图3 展示了一个量化卷积算子 与 TQC 之间的关系：

        ![图3. 量化 Conv 的执行过程](https://user-images.githubusercontent.com/43309460/190067258-4a4daec7-f898-4734-85a3-19f449c6a963.png)

        ### Quantize Delegate (量化代理函数)

        PPQ 允许你为网络中特定的 TQC 注册量化代理函数。这样你就可以注册自定义的量化处理逻辑，而非使用 PPQLinearQuantFunction 完成量化。

            def register_quantize_delegate(
                self, config: TensorQuantizationConfig,
                delegator: TorchQuantizeDelegator):

        使用 executor.register_quantize_delegate(config, function) 完成函数注册，被注册的函数必须满足 TorchQuantizeDelegator 所定义的接口。
        下面我们给出一个简单的量化代理函数例子：

            class MyQuantDelegator(TorchQuantizeDelegator):
                def __call__(self, tensor: torch.Tensor, config: TensorQuantizationConfig) -> torch.Tensor:
                    if config.policy.has_property(QuantizationProperty.ASYMMETRICAL):
                        raise ValueError('Sorry, this delegator handles only Symmetrical Quantizations.')
                    print('You are invoking cusitmized quant function now.')
                    return torch.round(tensor / config.scale) * config.scale

        在执行器遇到 TQC 时，将会调用 executor.quantize_function 执行量化，其逻辑为：

            def quantize_function(self, tensor: torch.Tensor, config: TensorQuantizationConfig = None) -> torch.Tensor:
                if config is None or not QuantizationStates.is_activated(config.state): return tensor
                elif config in self._delegates: return self._delegates[config](tensor, config)
                else:
                    if config.policy.has_property(QuantizationProperty.DYNAMIC):
                        return self._dynamic_quant_fn(tensor, config)
                    else:
                        return self._default_quant_fn(tensor, config)

        ### Usage (用法示例)

        PPQ 的执行器初始化需要一个计算图实例作为参数：

            executor = TorchExecutor(graph=ppq_quant_ir)
            executor.forward(inputs=..., output_names=..., hooks=...)

        这一计算图可以是量化过后的，也可以是没有量化的。但 PPQ 希望传入的计算图经过正确调度，传入没有调度的计算图将会触发警报：

            if not graph.extension_attrib.get(IS_DISPATCHED_GRAPH, False):
                ppq_warning('Can not create executor with your graph, graph is not correctly dispatched, '
                            'use dispatch_graph(graph=ir, platform=platfrom, setting=setting) first.')

        executor.forward 需要三个参数，下面举例对其进行说明：

            # 传入三个变量 a, b, c 作为输入
            executor.forward(inputs=[a, b, c], output_names=..., hooks=...)

            # 分别对图中 input, var 1 两个变量传入 a, b 作为输入
            executor.forward(inputs={'input': a, 'var 1': b}, output_names=..., hooks=...)

            # 传入一个完整的 tensor 作为输入
            executor.forward(inputs=torch.zeros(shape=[1,3,224,224]), output_names=..., hooks=...)

            # 要求网络输出 output, Var 1 的值
            executor.forward(inputs=..., output_names=['output 1', 'Var 1'], hooks=...)

        executor.forward 函数默认不需要梯度，如果希望执行带有梯度的网络，需要使用 executor.forward_with_gradient 函数。 forward 函数的返回值永远是一个 torch.Tensor 数组，其中元素的顺序由 output_names 参数决定。


        ### Hook (执行钩子函数)
        在调用 executor.forward 函数时可以传入 hooks 参数。钩子函数是注册在 op 上的，你可以传入一个字典用来说明需要调用的钩子函数：
        字典 {'Conv 1': myhook} 说明了希望在算子 Conv 1 的执行器件调用钩子函数 myhook。

        钩子函数必须继承于 RuntimeHook 类，必须实现成员函数 pre_forward_hook, post_forward_hook。在这两个函数中，你可以制定特定的逻辑修改算子输入输出的值。

            class RuntimeHook(metaclass=ABCMeta):
                def __init__(self, operation: Operation) -> None:
                    self._hook_to = operation

                def pre_forward_hook(self, inputs: list, **kwargs) -> list:
                    return inputs

                def post_forward_hook(self, outputs: list, **kwargs) -> list:
                    return outputs

            TorchExecutor - executor object which use torch as its backend.
                torch backend is used to graph simulating & training(QAT)

                all operation forward functions are written with pytorch,
                so that they will have gradient recorded by torch engine.

                which means you can directly access to tensor.grad after using output.backward()
        Args:
            graph (BaseGraph):
                executing graph object,
                TorchExecutor will automatically send all graph parameters towards executing device.

            fp16_mode (bool, optional): [whether the simulator is running in fp16 mode(unimplemented).]. Defaults to True.

            device (str, optional): [
                executing device, as same as torch.device,
                you can not select gpu to executing yet,
                graph will always be send to the very first visible cuda device.
            ]. Defaults to 'cuda'.
        """
        self._default_quant_fn = PPQuantFunction
        self._deployed = False
        self._device = device
        self._executing_context = TorchBackendContext(executing_device=self._device)
        super().__init__(graph)

        self._runnable_graph = RunnableGraph(self._graph)
        self._delegates = {}

        # fp16 is not available for now.
        self.fp16_mode = fp16_mode
        self.deploy()

    def register_quantize_delegate(self, config: TensorQuantizationConfig, delegator: TorchQuantizeDelegator):
        """Since PPQ 0.6.2, Interface TorchQuantizeDelegate is introduced to
        customize quantization logic: To be specific, you are suppose to
        inherit this class, and define your own computation logic within
        function __call__.

            Pass your Delegate to TorchExecutor by TorchExecutor.register_quantize_delegate(c, d)
                Where c is the target quantization config, d is your delegator class.
                Once you invoke this function, PPQ execution system will hand the quantization
                computation of config c over to your delegate. PPQ execution system will no
                longer quantize variable related with config c anymore.

        Notice that a delegate replaces quantization computation only, it still under the control of PPQ quantization
        System, so to say if your config has an invalid state like DEQUANTIZED, PPQ execution system will never been
        required to quantize related tensor and so your delegate class will take no effects on config c.

        Remove delegate function by TorchExecutor.remove_quantize_delegate(c)
        """
        if not isinstance(delegator, TorchQuantizeDelegator):
            raise TypeError(
                f'You can only register a TorchQuantizeDelegate as quantization delegator function,'
                f' however a/an {type(delegator)} was given'
            )
        if not isinstance(config, TensorQuantizationConfig):
            raise TypeError(f'Except a TensorQuantizationConfig instance, however {type(config)} was passed.')
        self._delegates[config] = delegator

    def remove_quantize_delegate(self, config: TensorQuantizationConfig):
        """Since PPQ 0.6.2, Interface TorchQuantizeDelegate is introduced to
        customize quantization logic: To be specific, you are suppose to
        inherit this class, and define your own computation logic within
        function __call__.

            Pass your Delegate to TorchExecutor by TorchExecutor.register_quantize_delegate(c, d)
                Where c is the target quantization config, d is your delegator class.
                Once you invoke this function, PPQ execution system will hand the quantization
                computation of config c over to your delegate. PPQ execution system will no
                longer quantize variable related with config c anymore.

        Notice that a delegate replaces quantization computation only, it still under the control of PPQ quantization
        System, so to say if your config has an invalid state like DEQUANTIZED, PPQ execution system will never been
        required to quantize related tensor and so your delegate class will take no effects on config c.

        Remove delegate function by TorchExecutor.remove_quantize_delegate(c)
        """
        if not isinstance(config, TensorQuantizationConfig):
            raise TypeError(f'Except a TensorQuantizationConfig instance, however {type(config)} was passed.')
        if config in self._delegates:
            self._delegates.pop(config)

    def deploy(self):
        """Deploy graph parameters towards target device.

        Raises:
            ValueError: [when target device is unacceptable]
        """
        self._deployed = True
        self._runnable_graph(GraphDeployCommand(device=self._device))

    def to(self, device: str):
        # just keep TorchExecutor behaving like torch.nn.Module
        self._device = torch.device(device)
        self.deploy()
        return self

    @torch.no_grad()
    def forward(
        self,
        inputs: Union[dict, list, torch.Tensor],
        output_names: List[str] = None,
        hooks: Dict[str, RuntimeHook] = None,
    ) -> List[torch.Tensor]:
        """Forward function of this executor.

        Notice this forward function will never store and compute gradients.

        Args:
            inputs (Union[dict, list, torch.Tensor]): [input tensor or somewhat]

            output_names (List[str], optional):
                onnx output node names, which used to confirm a output order.

                Defaults to None.

            hooks (Dict[str, RuntimeHook], optional):
                A hook table for customizing operation behaviour and collate data during executing.
                All hooks should inherit from class RuntimeHook, with all necessary methods implemented.
                    See also: esp_ppq.executor.base.RuntimeHook

                Executor calls hook.pre_forward_hook(operation, input_data) before dispatching operation,
                by using this feature, you can dynamically dispatch operation during executing,
                or processing input data as you want.(remember to return processed input data)

                Executor calls hook.post_forward_hook(operation, output_data) after the execution,
                you are supposed to  gather all necessary data from execution via this feature.

                For Quantable Operation, a much more powerful class:
                    esp_ppq.executor.base.QuantOpRuntimeHook is provided.
                see also: esp_ppq.executor.base.QuantOpRuntimeHook

                Defaults to None.

        Returns:
            List[torch.Tensor]: [executing result, list of tensor objects.]
        """
        return self.__forward(
            inputs=inputs, output_names=output_names, executing_order=self._executing_order, hooks=hooks
        )

    def forward_with_gradient(
        self,
        inputs: Union[dict, list, torch.Tensor],
        output_names: List[str] = None,
        hooks: Dict[str, RuntimeHook] = None,
    ) -> List[torch.Tensor]:
        """forward function of this executor.

            Notice this one will store and compute gradient.

        Args:
            inputs (Union[dict, list, torch.Tensor]): [input tensor or somewhat]
            output_names (List[str], optional):
                onnx output node names, which used to confirm a output order.

                Defaults to None.

            hooks (Dict[str, RuntimeHook], optional):
                A hook table for customizing operation behaviour and collate data during executing.
                All hooks should inherit from class RuntimeHook, with all necessary methods implemented.
                    See also: esp_ppq.executor.base.RuntimeHook

                Executor calls hook.pre_forward_hook(operation, input_data) before dispatching operation,
                by using this feature, you can dynamically dispatch operation during executing,
                or processing input data as you want.(remember to return processed input data)

                Executor calls hook.post_forward_hook(operation, output_data) after the execution,
                you are supposed to  gather all necessary data from execution via this feature.

                For Quantable Operation, a much more powerful class:
                    esp_ppq.executor.base.QuantOpRuntimeHook is provided.
                see also: esp_ppq.executor.base.QuantOpRuntimeHook

                Defaults to None.

        Returns:
            List[torch.Tensor]: [executing result, list of tensor objects.]
        """
        return self.__forward(
            inputs=inputs, output_names=output_names, executing_order=self._executing_order, hooks=hooks
        )

    def __forward(
        self,
        inputs: Union[dict, list, torch.Tensor],
        executing_order: List[Operation],
        output_names: List[str] = None,
        hooks: Dict[str, RuntimeHook] = None,
    ) -> List[torch.Tensor]:
        # processing with different input format
        if isinstance(inputs, dict):
            # directly feed value into variables
            for name, value in inputs.items():
                if name in self._graph.variables:
                    var = self._graph.variables[name]
                    var.value = value
                else:
                    print(f'Can not find variable {name} in your graph, please check.')
        else:
            inputs = self.prepare_input(inputs=inputs)
            for key, value in inputs.items():
                assert isinstance(value, torch.Tensor), (
                    f'TorchExecutor can only accept tensor as its input, while {type(value)} was given'
                )
                # input is acceptable, feed input value
                self._graph_input_dictionary[key].value = value

        # processing with output
        last_idx = 0  # record last variable
        if output_names is None:
            output_names = [name for name in self._graph.outputs]
        for name in output_names:
            if name not in self._graph.variables:
                raise KeyError(
                    f'You are requiring output value of variable {name}(is not a variable name), '
                    'however it is not a valid variable of current graph.'
                )
            source_op = self._graph.variables[name].source_op
            if source_op is not None:
                last_idx = max(last_idx, executing_order.index(source_op) + 1)

        visited_op, result_collector = [], [None for _ in output_names]
        # output name can be the same as input name, collect them directly.
        for name in output_names:
            if name in inputs:
                result_collector[output_names.index(name)] = inputs[name]

        for operation in executing_order[:last_idx]:
            try:
                assert isinstance(operation, Operation), 'Oops, seems you got something weird in your graph'
                assert isinstance(operation.platform, TargetPlatform), (
                    f'Operation {operation.name} has an invalid platform setting, '
                    f'only esp_ppq.core.TargetPlatform is expected here, while {type(operation.platform)} was given'
                )
                platform_dispatching_table = OPERATION_FORWARD_TABLE[operation.platform]
                if operation.type not in platform_dispatching_table:
                    raise NotImplementedError(
                        f'Graph op: {operation.name}({operation.type}) '
                        f'has no backend implementation on target platform {operation.platform}. '
                        'Register this op to esp_ppq.executor.base.py and esp_ppq.executor.op first'
                    )
                operation_forward_func = platform_dispatching_table[operation.type]
                operation_runtime_hook = (
                    hooks[operation.name] if (hooks is not None) and (operation.name in hooks) else None
                )
                inputs = [var.value for var in operation.inputs]

                # if operation is an QuantableOperation, we have to quant its inputs and outputs at first.
                if isinstance(operation, QuantableOperation):
                    input_configs = [_ for _ in operation.config.input_quantization_config]
                    if operation.platform in [
                        TargetPlatform.ESPDL_H_PRE_INT16,
                        TargetPlatform.ESPDL_S3_H_PRE_INT16,
                        TargetPlatform.ESPDL_C_H_PRE_INT16,
                    ]:
                        inputs = [
                            self.quantize_function(input, config).type(dtype=torch.float64)
                            if input is not None and input.dtype == torch.float32
                            else self.quantize_function(input, config)
                            for input, config in zip(inputs, input_configs)
                        ]
                    else:
                        inputs = [
                            self.quantize_function(input, config).type(dtype=torch.float32)
                            if input is not None and input.dtype == torch.float64
                            else self.quantize_function(input, config)
                            for input, config in zip(inputs, input_configs)
                        ]

                # PATCH 20220208
                # for idx, var in enumerate(operation.inputs):
                #     if var.name in output_names:
                #         result_collector[output_names.index(var.name)] = inputs[idx]

                # invoking pre-forward hook
                if operation_runtime_hook is not None:
                    if isinstance(operation_runtime_hook, QuantOPRuntimeHook):
                        inputs = operation_runtime_hook.pre_forward_hook(
                            inputs=[var.value for var in operation.inputs],
                            quant_inputs=inputs,
                            quant_configs=input_configs,
                        )
                    elif isinstance(operation_runtime_hook, RuntimeHook):
                        inputs = operation_runtime_hook.pre_forward_hook(inputs=inputs)
                    else:
                        raise TypeError(f'invalid hook instance was given with operation: {operation}')

                # forward and collecting result
                outputs = operation_forward_func(operation, inputs, self._executing_context)
                outputs = outputs if isinstance(outputs, (list, tuple)) else [outputs]
                fp_outputs = outputs

                # quantize all result if is necessary
                if isinstance(operation, QuantableOperation):
                    output_configs = [_ for _ in operation.config.output_quantization_config]
                    if operation.platform in [
                        TargetPlatform.ESPDL_H_PRE_INT16,
                        TargetPlatform.ESPDL_S3_H_PRE_INT16,
                        TargetPlatform.ESPDL_C_H_PRE_INT16,
                    ]:
                        outputs = [
                            self.quantize_function(output, config).type(dtype=torch.float64)
                            if output is not None and output.dtype == torch.float32
                            else self.quantize_function(output, config)
                            for output, config in zip(outputs, output_configs)
                        ]
                    else:
                        outputs = [
                            self.quantize_function(output, config).type(dtype=torch.float32)
                            if output is not None and output.dtype == torch.float64
                            else self.quantize_function(output, config)
                            for output, config in zip(outputs, output_configs)
                        ]

                # invoking post-forward hook
                if operation_runtime_hook is not None:
                    if isinstance(operation_runtime_hook, QuantOPRuntimeHook):
                        outputs = operation_runtime_hook.post_forward_hook(
                            outputs=fp_outputs, quant_outputs=outputs, quant_configs=output_configs
                        )
                    elif isinstance(operation_runtime_hook, RuntimeHook):
                        outputs = operation_runtime_hook.post_forward_hook(outputs=outputs)
                    else:
                        raise TypeError(f'invalid hook instance was given with operation: {operation}')

                # feed value to graph variables.
                for output_idx, output_var in enumerate(operation.outputs):
                    output_var = operation.outputs[output_idx]
                    output_var.value = outputs[output_idx]

                    if output_var.name in output_names:
                        result_collector[output_names.index(output_var.name)] = outputs[output_idx]
            except Exception as _:
                raise RuntimeError(f'Op Execution Error: {str(operation)}') from _

            # remove useless value(runtime clear).
            visited_op.append(operation)
            for var in operation.inputs:
                if var.is_parameter:
                    continue
                if all(op in visited_op for op in var.dest_ops):
                    var.value = None

        # clear all variable(static clear).
        for var in self._graph.variables.values():
            if not var.is_parameter:
                var.value = None
        # end for
        return result_collector

    @torch.no_grad()
    @empty_ppq_cache
    def tracing_operation_meta(
        self,
        inputs: Union[dict, list, torch.Tensor],
        output_names: List[str] = None,
    ) -> None:
        """Tracing meta data for each operation, if there are some already
        created meta data with your operation, They will be override without
        warrning.

        Args:
            inputs (Union[dict, list, torch.Tensor]): [description]
            output_names (List[str], optional): [description]. Defaults to None.
        """
        hooks = {}
        for op_name, operation in self._graph.operations.items():
            hooks[op_name] = TorchMetaDataTracingHook(operation=operation)

        self.__forward(inputs=inputs, output_names=output_names, executing_order=self._executing_order, hooks=hooks)

    def load_graph(self, graph: BaseGraph) -> dict:
        super().load_graph(graph)
        self._deployed = False
        self._runnable_graph = RunnableGraph(self._graph)
        self._runnable_graph(GraphDeployCommand(device=self._device))

    def quantize_function(self, tensor: torch.Tensor, config: TensorQuantizationConfig = None) -> torch.Tensor:
        if config is None or not QuantizationStates.is_activated(config.state):
            return tensor
        elif config in self._delegates:
            return self._delegates[config](tensor, config)
        else:
            return self._default_quant_fn(tensor, config)

    def dummy_forward(self, hooks: Dict[str, RuntimeHook] = None) -> None:
        """This function allows you to execute entire graph without feeding any
        data. This feature is required for operation parameter quantization.
        See also: esp_ppq.quantization.optim.ParameterQuantizePass.

        This function fakes some input tensors via operation metadata.
            ATTENTION: operation must have metadata before invoking this function.

        Args:
            hooks (Dict[str, RuntimeHook], optional):
                A hook table for customizing operation behaviour and collate data during executing.
                All hooks should inherit from class RuntimeHook, with all necessary methods implemented.
                    See also: esp_ppq.executor.base.RuntimeHook

                Executor calls hook.pre_forward_hook(operation, input_data) before dispatching operation,
                by using this feature, you can dynamically dispatch operation during executing,
                or processing input data as you want.(remember to return processed input data)

                Executor calls hook.post_forward_hook(operation, output_data) after the execution,
                you are supposed to  gather all necessary data from execution via this feature.

                For Quantable Operation, a much more powerful class:
                    esp_ppq.executor.base.QuantOpRuntimeHook is provided.
                see also: esp_ppq.executor.base.QuantOpRuntimeHook

                Defaults to None.
        """
        # build dummy input based on meta data
        feed_dict = {}
        for var_name, input_var in self._graph.inputs.items():
            if len(input_var.dest_ops) == 0:
                continue

            assert input_var.shape is not None, (
                f'Can not generate dummy input for input variable {input_var.name}, input shape is not specified.'
            )

            feed_dict[var_name] = (
                torch.Tensor(size=input_var.shape, device='cpu')
                .fill_(0)
                .type(dtype=DataType.to_torch(input_var.dtype))
                .to(self._device)
            )
        self.forward(inputs=feed_dict, hooks=hooks)

    def partial_graph_forward(
        self, operations: List[Operation], feed_dict: Dict[str, torch.Tensor], output_names: List[str]
    ) -> List[torch.Tensor]:
        """This forward function allows you to execute a series operations in
        your graph. (only operations list in your params will be executed with
        this function) Which serves as a great feature for quantization aware
        training.

        Args:
            operations (List[Operation]):
                operations that you want to execute,
                notice that executing will strictly follow your operation order.

            feed_dict (Dict[str, torch.Tensor]):
                an dictionary contains {variable name: value}, as an input to this execution.

            output_names (List[str]):
                output variable names.

        Returns:
            List[torch.Tensor]: [description]
        """

        return self.__forward(inputs=feed_dict, output_names=output_names, executing_order=operations)



import torchvision
from Utilities.Imagenet import (
    evaluate_mmlab_module_with_imagenet,
    evaluate_onnx_module_with_imagenet,
    evaluate_ppq_module_with_imagenet,
    evaluate_torch_module_with_imagenet,
    load_imagenet_from_directory,
)

from esp_ppq import *
from esp_ppq.api import *

"""
    使用这个脚本来测试量化 torchvision 中的典型分类模型
        使用 imagenet 中的数据测试量化精度与 calibration
        默认的 imagenet 数据集位置: Assets/Imagenet_Train, Assets/Imagenet_Valid
        你可以通过软连接创建它们:
            ln -s /home/data/Imagenet/val Assets/Imagenet_Valid
            ln -s /home/data/Imagenet/train Assets/Imagenet_Train
"""

CFG_DEVICE = 'cuda'  # 一个神奇的字符串，用来确定执行设备
CFG_BATCHSIZE = 64  # 测试与calib时的 batchsize
CFG_INPUT_SHAPE = (CFG_BATCHSIZE, 3, 224, 224)  # 用来确定模型输入的尺寸，好像 imagenet 都是这个尺寸
CFG_VALIDATION_DIR = 'Assets/Imagenet_Valid'  # 用来读取 validation dataset
CFG_TRAIN_DIR = 'Assets/Imagenet_Train'  # 用来读取 train dataset，注意该集合将被用来 calibrate 你的模型
CFG_PLATFORM = TargetPlatform.PPL_CUDA_INT8  # 用来指定目标平台
CFG_DUMP_PATH = 'Output/'  # 所有模型保存的路径名
QUANT_SETTING = QuantizationSettingFactory.default_setting()  # 用来指定量化配置
QUANT_SETTING.lsq_optimization = True

with ENABLE_CUDA_KERNEL():
    if __name__ == '__main__':
        for model_builder, model_name in (
            # (torchvision.models.alexnet, 'alexnet'),
            (torchvision.models.inception_v3, 'inception_v3'),
            (torchvision.models.mnasnet0_5, 'mnasnet0_5'),
            (torchvision.models.mnasnet1_0, 'mnasnet1_0'),
            (torchvision.models.squeezenet1_0, 'squeezenet1_0'),
            # (torchvision.models.squeezenet1_1, 'squeezenet1_1'),
            (torchvision.models.shufflenet_v2_x1_0, 'shufflenet_v2_x1_0'),
            (torchvision.models.resnet18, 'resnet18'),
            (torchvision.models.resnet34, 'resnet34'),
            (torchvision.models.resnet50, 'resnet50'),
            # (torchvision.models.resnext50_32x4d, 'resnext50_32x4d'),
            # (torchvision.models.wide_resnet50_2, 'wide_resnet50_2'),
            (torchvision.models.googlenet, 'googlenet'),
            (torchvision.models.vgg11, 'vgg11'),
            # (torchvision.models.vgg16, 'vgg16'),
            (torchvision.models.mobilenet.mobilenet_v2, 'mobilenet_v2'),
        ):
            print(f'---------------------- PPQ Quantization Test Running with {model_name} ----------------------')
            model = model_builder(pretrained=True).to(CFG_DEVICE)

            dataloader = load_imagenet_from_directory(
                directory=CFG_TRAIN_DIR,
                batchsize=CFG_BATCHSIZE,
                shuffle=False,
                subset=5120,
                require_label=False,
                num_of_workers=8,
            )

            ppq_quant_ir = quantize_torch_model(
                model=model,
                calib_dataloader=dataloader,
                input_shape=CFG_INPUT_SHAPE,
                calib_steps=5120 // CFG_BATCHSIZE,
                collate_fn=lambda x: x.to(CFG_DEVICE),
                verbose=1,
                device=CFG_DEVICE,
                platform=CFG_PLATFORM,
                setting=QUANT_SETTING,
            )

            ppq_int8_report = evaluate_ppq_module_with_imagenet(
                model=ppq_quant_ir,
                imagenet_validation_dir=CFG_VALIDATION_DIR,
                batchsize=CFG_BATCHSIZE,
                device=CFG_DEVICE,
                verbose=True,
            )

            export_ppq_graph(
                graph=ppq_quant_ir,
                platform=TargetPlatform.ONNXRUNTIME,
                graph_save_to=f'{os.path.join(CFG_DUMP_PATH, model_name)}.onnx',
            )

            evaluate_onnx_module_with_imagenet(
                onnxruntime_model_path=f'{os.path.join(CFG_DUMP_PATH, model_name)}.onnx',
                imagenet_validation_dir=CFG_VALIDATION_DIR,
                batchsize=CFG_BATCHSIZE,
                device=CFG_DEVICE,
            )
    else:
        raise Exception('You may not import this file.')



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Tensor(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Tensor()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTensor(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Tensor
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Tensor
    def Dims(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Tensor
    def DimsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int64Flags, o)
        return 0

    # Tensor
    def DimsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def DimsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Tensor
    def DataType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Tensor
    def FloatData(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Tensor
    def FloatDataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # Tensor
    def FloatDataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def FloatDataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Tensor
    def Int32Data(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Tensor
    def Int32DataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # Tensor
    def Int32DataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def Int32DataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Tensor
    def StringData(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Tensor
    def StringDataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def StringDataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Tensor
    def Int64Data(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Tensor
    def Int64DataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int64Flags, o)
        return 0

    # Tensor
    def Int64DataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def Int64DataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Tensor
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Tensor
    def DocString(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Tensor
    def RawData(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 16
            from FlatBuffers.Dl.AlignedBytes import AlignedBytes
            obj = AlignedBytes()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Tensor
    def RawDataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def RawDataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        return o == 0

    # Tensor
    def ExternalData(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.StringStringEntry import StringStringEntry
            obj = StringStringEntry()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Tensor
    def ExternalDataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def ExternalDataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        return o == 0

    # Tensor
    def DataLocation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Tensor
    def DoubleData(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Tensor
    def DoubleDataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float64Flags, o)
        return 0

    # Tensor
    def DoubleDataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def DoubleDataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        return o == 0

    # Tensor
    def Uint64Data(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Tensor
    def Uint64DataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # Tensor
    def Uint64DataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def Uint64DataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        return o == 0

    # Tensor
    def Exponents(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Tensor
    def ExponentsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int64Flags, o)
        return 0

    # Tensor
    def ExponentsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Tensor
    def ExponentsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        return o == 0

def TensorStart(builder):
    builder.StartObject(14)

def Start(builder):
    TensorStart(builder)

def TensorAddDims(builder, dims):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(dims), 0)

def AddDims(builder, dims):
    TensorAddDims(builder, dims)

def TensorStartDimsVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)

def StartDimsVector(builder, numElems):
    return TensorStartDimsVector(builder, numElems)

def TensorAddDataType(builder, dataType):
    builder.PrependInt32Slot(1, dataType, 0)

def AddDataType(builder, dataType):
    TensorAddDataType(builder, dataType)

def TensorAddFloatData(builder, floatData):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(floatData), 0)

def AddFloatData(builder, floatData):
    TensorAddFloatData(builder, floatData)

def TensorStartFloatDataVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartFloatDataVector(builder, numElems):
    return TensorStartFloatDataVector(builder, numElems)

def TensorAddInt32Data(builder, int32Data):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(int32Data), 0)

def AddInt32Data(builder, int32Data):
    TensorAddInt32Data(builder, int32Data)

def TensorStartInt32DataVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartInt32DataVector(builder, numElems):
    return TensorStartInt32DataVector(builder, numElems)

def TensorAddStringData(builder, stringData):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(stringData), 0)

def AddStringData(builder, stringData):
    TensorAddStringData(builder, stringData)

def TensorStartStringDataVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartStringDataVector(builder, numElems):
    return TensorStartStringDataVector(builder, numElems)

def TensorAddInt64Data(builder, int64Data):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(int64Data), 0)

def AddInt64Data(builder, int64Data):
    TensorAddInt64Data(builder, int64Data)

def TensorStartInt64DataVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)

def StartInt64DataVector(builder, numElems):
    return TensorStartInt64DataVector(builder, numElems)

def TensorAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)

def AddName(builder, name):
    TensorAddName(builder, name)

def TensorAddDocString(builder, docString):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(docString), 0)

def AddDocString(builder, docString):
    TensorAddDocString(builder, docString)

def TensorAddRawData(builder, rawData):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(rawData), 0)

def AddRawData(builder, rawData):
    TensorAddRawData(builder, rawData)

def TensorStartRawDataVector(builder, numElems):
    return builder.StartVector(16, numElems, 16)

def StartRawDataVector(builder, numElems):
    return TensorStartRawDataVector(builder, numElems)

def TensorAddExternalData(builder, externalData):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(externalData), 0)

def AddExternalData(builder, externalData):
    TensorAddExternalData(builder, externalData)

def TensorStartExternalDataVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartExternalDataVector(builder, numElems):
    return TensorStartExternalDataVector(builder, numElems)

def TensorAddDataLocation(builder, dataLocation):
    builder.PrependInt32Slot(10, dataLocation, 0)

def AddDataLocation(builder, dataLocation):
    TensorAddDataLocation(builder, dataLocation)

def TensorAddDoubleData(builder, doubleData):
    builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(doubleData), 0)

def AddDoubleData(builder, doubleData):
    TensorAddDoubleData(builder, doubleData)

def TensorStartDoubleDataVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)

def StartDoubleDataVector(builder, numElems):
    return TensorStartDoubleDataVector(builder, numElems)

def TensorAddUint64Data(builder, uint64Data):
    builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(uint64Data), 0)

def AddUint64Data(builder, uint64Data):
    TensorAddUint64Data(builder, uint64Data)

def TensorStartUint64DataVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)

def StartUint64DataVector(builder, numElems):
    return TensorStartUint64DataVector(builder, numElems)

def TensorAddExponents(builder, exponents):
    builder.PrependUOffsetTRelativeSlot(13, flatbuffers.number_types.UOffsetTFlags.py_type(exponents), 0)

def AddExponents(builder, exponents):
    TensorAddExponents(builder, exponents)

def TensorStartExponentsVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)

def StartExponentsVector(builder, numElems):
    return TensorStartExponentsVector(builder, numElems)

def TensorEnd(builder):
    return builder.EndObject()

def End(builder):
    return TensorEnd(builder)

import FlatBuffers.Dl.AlignedBytes
import FlatBuffers.Dl.StringStringEntry
try:
    from typing import List
except:
    pass

class TensorT(object):

    # TensorT
    def __init__(self):
        self.dims = None  # type: List[int]
        self.dataType = 0  # type: int
        self.floatData = None  # type: List[float]
        self.int32Data = None  # type: List[int]
        self.stringData = None  # type: List[str]
        self.int64Data = None  # type: List[int]
        self.name = None  # type: str
        self.docString = None  # type: str
        self.rawData = None  # type: List[FlatBuffers.Dl.AlignedBytes.AlignedBytesT]
        self.externalData = None  # type: List[FlatBuffers.Dl.StringStringEntry.StringStringEntryT]
        self.dataLocation = 0  # type: int
        self.doubleData = None  # type: List[float]
        self.uint64Data = None  # type: List[int]
        self.exponents = None  # type: List[int]

    @classmethod
    def InitFromBuf(cls, buf, pos):
        tensor = Tensor()
        tensor.Init(buf, pos)
        return cls.InitFromObj(tensor)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, tensor):
        x = TensorT()
        x._UnPack(tensor)
        return x

    # TensorT
    def _UnPack(self, tensor):
        if tensor is None:
            return
        if not tensor.DimsIsNone():
            if np is None:
                self.dims = []
                for i in range(tensor.DimsLength()):
                    self.dims.append(tensor.Dims(i))
            else:
                self.dims = tensor.DimsAsNumpy()
        self.dataType = tensor.DataType()
        if not tensor.FloatDataIsNone():
            if np is None:
                self.floatData = []
                for i in range(tensor.FloatDataLength()):
                    self.floatData.append(tensor.FloatData(i))
            else:
                self.floatData = tensor.FloatDataAsNumpy()
        if not tensor.Int32DataIsNone():
            if np is None:
                self.int32Data = []
                for i in range(tensor.Int32DataLength()):
                    self.int32Data.append(tensor.Int32Data(i))
            else:
                self.int32Data = tensor.Int32DataAsNumpy()
        if not tensor.StringDataIsNone():
            self.stringData = []
            for i in range(tensor.StringDataLength()):
                self.stringData.append(tensor.StringData(i))
        if not tensor.Int64DataIsNone():
            if np is None:
                self.int64Data = []
                for i in range(tensor.Int64DataLength()):
                    self.int64Data.append(tensor.Int64Data(i))
            else:
                self.int64Data = tensor.Int64DataAsNumpy()
        self.name = tensor.Name()
        self.docString = tensor.DocString()
        if not tensor.RawDataIsNone():
            self.rawData = []
            for i in range(tensor.RawDataLength()):
                if tensor.RawData(i) is None:
                    self.rawData.append(None)
                else:
                    alignedBytes_ = FlatBuffers.Dl.AlignedBytes.AlignedBytesT.InitFromObj(tensor.RawData(i))
                    self.rawData.append(alignedBytes_)
        if not tensor.ExternalDataIsNone():
            self.externalData = []
            for i in range(tensor.ExternalDataLength()):
                if tensor.ExternalData(i) is None:
                    self.externalData.append(None)
                else:
                    stringStringEntry_ = FlatBuffers.Dl.StringStringEntry.StringStringEntryT.InitFromObj(tensor.ExternalData(i))
                    self.externalData.append(stringStringEntry_)
        self.dataLocation = tensor.DataLocation()
        if not tensor.DoubleDataIsNone():
            if np is None:
                self.doubleData = []
                for i in range(tensor.DoubleDataLength()):
                    self.doubleData.append(tensor.DoubleData(i))
            else:
                self.doubleData = tensor.DoubleDataAsNumpy()
        if not tensor.Uint64DataIsNone():
            if np is None:
                self.uint64Data = []
                for i in range(tensor.Uint64DataLength()):
                    self.uint64Data.append(tensor.Uint64Data(i))
            else:
                self.uint64Data = tensor.Uint64DataAsNumpy()
        if not tensor.ExponentsIsNone():
            if np is None:
                self.exponents = []
                for i in range(tensor.ExponentsLength()):
                    self.exponents.append(tensor.Exponents(i))
            else:
                self.exponents = tensor.ExponentsAsNumpy()

    # TensorT
    def Pack(self, builder):
        if self.dims is not None:
            if np is not None and type(self.dims) is np.ndarray:
                dims = builder.CreateNumpyVector(self.dims)
            else:
                TensorStartDimsVector(builder, len(self.dims))
                for i in reversed(range(len(self.dims))):
                    builder.PrependInt64(self.dims[i])
                dims = builder.EndVector()
        if self.floatData is not None:
            if np is not None and type(self.floatData) is np.ndarray:
                floatData = builder.CreateNumpyVector(self.floatData)
            else:
                TensorStartFloatDataVector(builder, len(self.floatData))
                for i in reversed(range(len(self.floatData))):
                    builder.PrependFloat32(self.floatData[i])
                floatData = builder.EndVector()
        if self.int32Data is not None:
            if np is not None and type(self.int32Data) is np.ndarray:
                int32Data = builder.CreateNumpyVector(self.int32Data)
            else:
                TensorStartInt32DataVector(builder, len(self.int32Data))
                for i in reversed(range(len(self.int32Data))):
                    builder.PrependInt32(self.int32Data[i])
                int32Data = builder.EndVector()
        if self.stringData is not None:
            stringDatalist = []
            for i in range(len(self.stringData)):
                stringDatalist.append(builder.CreateString(self.stringData[i]))
            TensorStartStringDataVector(builder, len(self.stringData))
            for i in reversed(range(len(self.stringData))):
                builder.PrependUOffsetTRelative(stringDatalist[i])
            stringData = builder.EndVector()
        if self.int64Data is not None:
            if np is not None and type(self.int64Data) is np.ndarray:
                int64Data = builder.CreateNumpyVector(self.int64Data)
            else:
                TensorStartInt64DataVector(builder, len(self.int64Data))
                for i in reversed(range(len(self.int64Data))):
                    builder.PrependInt64(self.int64Data[i])
                int64Data = builder.EndVector()
        if self.name is not None:
            name = builder.CreateString(self.name)
        if self.docString is not None:
            docString = builder.CreateString(self.docString)
        if self.rawData is not None:
            TensorStartRawDataVector(builder, len(self.rawData))
            for i in reversed(range(len(self.rawData))):
                self.rawData[i].Pack(builder)
            rawData = builder.EndVector()
        if self.externalData is not None:
            externalDatalist = []
            for i in range(len(self.externalData)):
                externalDatalist.append(self.externalData[i].Pack(builder))
            TensorStartExternalDataVector(builder, len(self.externalData))
            for i in reversed(range(len(self.externalData))):
                builder.PrependUOffsetTRelative(externalDatalist[i])
            externalData = builder.EndVector()
        if self.doubleData is not None:
            if np is not None and type(self.doubleData) is np.ndarray:
                doubleData = builder.CreateNumpyVector(self.doubleData)
            else:
                TensorStartDoubleDataVector(builder, len(self.doubleData))
                for i in reversed(range(len(self.doubleData))):
                    builder.PrependFloat64(self.doubleData[i])
                doubleData = builder.EndVector()
        if self.uint64Data is not None:
            if np is not None and type(self.uint64Data) is np.ndarray:
                uint64Data = builder.CreateNumpyVector(self.uint64Data)
            else:
                TensorStartUint64DataVector(builder, len(self.uint64Data))
                for i in reversed(range(len(self.uint64Data))):
                    builder.PrependUint64(self.uint64Data[i])
                uint64Data = builder.EndVector()
        if self.exponents is not None:
            if np is not None and type(self.exponents) is np.ndarray:
                exponents = builder.CreateNumpyVector(self.exponents)
            else:
                TensorStartExponentsVector(builder, len(self.exponents))
                for i in reversed(range(len(self.exponents))):
                    builder.PrependInt64(self.exponents[i])
                exponents = builder.EndVector()
        TensorStart(builder)
        if self.dims is not None:
            TensorAddDims(builder, dims)
        TensorAddDataType(builder, self.dataType)
        if self.floatData is not None:
            TensorAddFloatData(builder, floatData)
        if self.int32Data is not None:
            TensorAddInt32Data(builder, int32Data)
        if self.stringData is not None:
            TensorAddStringData(builder, stringData)
        if self.int64Data is not None:
            TensorAddInt64Data(builder, int64Data)
        if self.name is not None:
            TensorAddName(builder, name)
        if self.docString is not None:
            TensorAddDocString(builder, docString)
        if self.rawData is not None:
            TensorAddRawData(builder, rawData)
        if self.externalData is not None:
            TensorAddExternalData(builder, externalData)
        TensorAddDataLocation(builder, self.dataLocation)
        if self.doubleData is not None:
            TensorAddDoubleData(builder, doubleData)
        if self.uint64Data is not None:
            TensorAddUint64Data(builder, uint64Data)
        if self.exponents is not None:
            TensorAddExponents(builder, exponents)
        tensor = TensorEnd(builder)
        return tensor



from collections import defaultdict
from typing import Callable, Dict, Iterable, List, Union

import torch
from tqdm import tqdm

from esp_ppq.core import convert_any_to_numpy
from esp_ppq.executor import TorchExecutor
from esp_ppq.IR import BaseGraph, QuantableOperation
from esp_ppq.IR.quantize import QuantableGraph
from esp_ppq.utils.fetch import tensor_random_fetch

from .util import MeasurePrinter, MeasureRecorder


def layerwise_error_analyse(
    graph: BaseGraph,
    dataloader: Iterable,
    interested_outputs: Union[str, List[str]] = None,
    collate_fn: Callable = None,
    running_device='cuda',
    method: str = 'snr',
    steps: int = 8,
    verbose: bool = True,
) -> Dict[str, tuple]:
    """Measure the quantization error of each operation A dictionary contains
    output differences for all operation will be returned as a result.

        Result is like: {'operation name 1': 0.933, 'operation name 2': 0.926}

    if verbose is set as True, this function will display error report at last.

    The key of the dictionary is an operation name while the value of corresponding key
        is the difference between quantized output and float output of this operation.

    Result {'operation name 1': 0.933} means quantizing operation 1
        will generates 0.933 quantization error to output variable

    ATTENTION: Output difference is measured at operation-level.

    Args:
        graph (BaseGraph):
            A fully quantized graph instance.

        running_device (str):
            A device string used to initialize a graph executor for the graph execution.
                if a executor was given, this parameter will be skipped.

        dataloader (Iterator):
            Test dataloader, this function will measure quantization error based on given data.

        collate_fn (Callable, optional):
            An data preprocessing function provided by user to convert data from dataloader towards
                executable format. If set as None, then no action will be taken during preprocessing.

        method (str, optional):
            A string indicates a measurement to calculate the difference of quantized output and fp32 one.
                'cosine', 'snr', and 'mse' is supported in PPQ for now.

        steps (Int, optional)
            computation steps.

        interested_outputs (Union[str, List[str]] = None)
            a list contains your interested output variables.
                if set as None, all graph output variables will be measured via this function.

    Returns:
        A dictionary contains output differences for all operation will be returned from this function.

        Result is like: {'operation name 1': 0.933, 'operation name 2': 0.926}
    """

    if interested_outputs is None:
        interested_outputs = [name for name in graph.outputs]

    if isinstance(interested_outputs, str):
        interested_outputs = [interested_outputs]

    executor = TorchExecutor(graph=graph, device=running_device)

    # find all quantable operations.
    quantable_operations = [
        operation
        for operation in graph.operations.values()
        if (isinstance(operation, QuantableOperation) and operation.is_computing_op)
    ]
    if len(quantable_operations) == 0:
        print('Oops. you got nothing to analyse.')
        return

    for operation in graph.operations.values():
        if isinstance(operation, QuantableOperation):
            operation.dequantize()

            # # we only need reports from computing op.
            # if operation.is_computing_op:
            #     quantable_operations.append(operation)

    # dequantize all operations, create recorder for each operation
    recorders = {}
    for operation in quantable_operations:
        if isinstance(operation, QuantableOperation):
            recorders[operation.name] = MeasureRecorder(measurement=method)

    # run for each quantable operations:
    for operation in tqdm(quantable_operations, desc='Analysing Layerwise quantization error:'):
        assert isinstance(operation, QuantableOperation)
        recorder = recorders[operation.name]
        assert isinstance(recorder, MeasureRecorder)

        for idx, batch in enumerate(dataloader):
            if collate_fn is not None:
                batch = collate_fn(batch)
            fp_outputs = executor.forward(inputs=batch, output_names=interested_outputs)

            # manually override quantization state
            operation.restore_quantize_state()
            qt_outputs = executor.forward(inputs=batch, output_names=interested_outputs)

            for fp_output, qt_output in zip(fp_outputs, qt_outputs):
                recorder.update(y_pred=qt_output, y_real=fp_output)

            # manually override quantization state
            operation.dequantize()
            if idx >= steps:
                break

    # restore quantization states
    for operation in graph.operations.values():
        if isinstance(operation, QuantableOperation):
            operation.restore_quantize_state()

    results = {}
    for operation in quantable_operations:
        assert isinstance(operation, QuantableOperation)
        results[operation.name] = recorders[operation.name].measure

    if verbose:
        method_str = 'MEASUREMENT'
        if method == 'snr':
            method_str = 'NOISE:SIGNAL POWER RATIO'
        if method == 'cosine':
            method_str = 'COSINE SIMILARITY'
        if method == 'mse':
            method_str = 'MSE LOSS(UNSCALED)'
        MeasurePrinter(
            results, order='large_to_small', measure=method_str, percentage=method in {'snr', 'cosine'}
        ).print()
    return results


def variable_analyse(
    graph: BaseGraph,
    dataloader: Iterable,
    interested_outputs: Union[str, List[str]],
    collate_fn: Callable = None,
    running_device='cuda',
    samples_per_step: int = 65536,
    steps: int = 8,
    dequantize: bool = False,
):
    quant_graph = QuantableGraph(graph)

    executor = TorchExecutor(graph=graph, device=running_device)
    if dequantize:
        quant_graph.dequantize_graph()

    data_collector = defaultdict(list)
    for idx, batch in enumerate(dataloader):
        if collate_fn is not None:
            batch = collate_fn(batch)
        fp_outputs = executor.forward(inputs=batch, output_names=interested_outputs)
        for output, output_name in zip(fp_outputs, interested_outputs):
            data_collector[output_name].append(
                tensor_random_fetch(tensor=output, num_of_fetches=samples_per_step).unsqueeze(0)
            )
        if idx >= steps:
            break

    for name in interested_outputs:
        tensor = torch.cat(data_collector[name]).flatten()
        tensor = convert_any_to_numpy(tensor)

        try:
            from matplotlib import pyplot as plt
        except ImportError as e:
            raise Exception('Install matplotlib before using this function.')

        plt.figure(figsize=[12, 8])
        plt.title(f'Histogram Result of Variable {name}:')
        plt.hist(tensor, bins=64)
        plt.show()

    if dequantize:
        quant_graph.restore_quantize_state()


def parameter_analyse(graph: BaseGraph):
    ranges, stds, means = {}, {}, {}
    for operation in graph.operations.values():
        for var in operation.parameters:
            value = var.value
            assert isinstance(value, torch.Tensor), (
                f'Invaild parameter value type, expect torch.Tensor, however {type(value)} was given.'
            )
            if value.numel() <= 1:
                continue

            _min, _max, _std, _mean = 0, 0, 0, 0
            try:
                _min = value.min().item()
                _max = value.max().item()
                _std = value.std().item()
                _mean = value.mean().item()
            except:
                pass

            ranges[f'{var.name}[{operation.name}]'] = _max - _min
            stds[f'{var.name}[{operation.name}]'] = _std
            means[f'{var.name}[{operation.name}]'] = abs(_mean)

    MeasurePrinter(ranges, order='large_to_small', measure='Value Range').print()
    MeasurePrinter(stds, order='large_to_small', measure='Value Std').print()
    MeasurePrinter(means, order='large_to_small', measure='Value Mean(Abs)').print()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Node(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Node()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsNode(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Node
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Node
    def Input(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Node
    def InputLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Node
    def InputIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Node
    def Output(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Node
    def OutputLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Node
    def OutputIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Node
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Node
    def OpType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Node
    def Domain(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Node
    def Attribute(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.Attribute import Attribute
            obj = Attribute()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Node
    def AttributeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Node
    def AttributeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Node
    def DocString(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def NodeStart(builder):
    builder.StartObject(7)

def Start(builder):
    NodeStart(builder)

def NodeAddInput(builder, input):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(input), 0)

def AddInput(builder, input):
    NodeAddInput(builder, input)

def NodeStartInputVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartInputVector(builder, numElems):
    return NodeStartInputVector(builder, numElems)

def NodeAddOutput(builder, output):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(output), 0)

def AddOutput(builder, output):
    NodeAddOutput(builder, output)

def NodeStartOutputVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartOutputVector(builder, numElems):
    return NodeStartOutputVector(builder, numElems)

def NodeAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)

def AddName(builder, name):
    NodeAddName(builder, name)

def NodeAddOpType(builder, opType):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(opType), 0)

def AddOpType(builder, opType):
    NodeAddOpType(builder, opType)

def NodeAddDomain(builder, domain):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(domain), 0)

def AddDomain(builder, domain):
    NodeAddDomain(builder, domain)

def NodeAddAttribute(builder, attribute):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(attribute), 0)

def AddAttribute(builder, attribute):
    NodeAddAttribute(builder, attribute)

def NodeStartAttributeVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartAttributeVector(builder, numElems):
    return NodeStartAttributeVector(builder, numElems)

def NodeAddDocString(builder, docString):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(docString), 0)

def AddDocString(builder, docString):
    NodeAddDocString(builder, docString)

def NodeEnd(builder):
    return builder.EndObject()

def End(builder):
    return NodeEnd(builder)

import FlatBuffers.Dl.Attribute
try:
    from typing import List
except:
    pass

class NodeT(object):

    # NodeT
    def __init__(self):
        self.input = None  # type: List[str]
        self.output = None  # type: List[str]
        self.name = None  # type: str
        self.opType = None  # type: str
        self.domain = None  # type: str
        self.attribute = None  # type: List[FlatBuffers.Dl.Attribute.AttributeT]
        self.docString = None  # type: str

    @classmethod
    def InitFromBuf(cls, buf, pos):
        node = Node()
        node.Init(buf, pos)
        return cls.InitFromObj(node)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, node):
        x = NodeT()
        x._UnPack(node)
        return x

    # NodeT
    def _UnPack(self, node):
        if node is None:
            return
        if not node.InputIsNone():
            self.input = []
            for i in range(node.InputLength()):
                self.input.append(node.Input(i))
        if not node.OutputIsNone():
            self.output = []
            for i in range(node.OutputLength()):
                self.output.append(node.Output(i))
        self.name = node.Name()
        self.opType = node.OpType()
        self.domain = node.Domain()
        if not node.AttributeIsNone():
            self.attribute = []
            for i in range(node.AttributeLength()):
                if node.Attribute(i) is None:
                    self.attribute.append(None)
                else:
                    attribute_ = FlatBuffers.Dl.Attribute.AttributeT.InitFromObj(node.Attribute(i))
                    self.attribute.append(attribute_)
        self.docString = node.DocString()

    # NodeT
    def Pack(self, builder):
        if self.input is not None:
            inputlist = []
            for i in range(len(self.input)):
                inputlist.append(builder.CreateString(self.input[i]))
            NodeStartInputVector(builder, len(self.input))
            for i in reversed(range(len(self.input))):
                builder.PrependUOffsetTRelative(inputlist[i])
            input = builder.EndVector()
        if self.output is not None:
            outputlist = []
            for i in range(len(self.output)):
                outputlist.append(builder.CreateString(self.output[i]))
            NodeStartOutputVector(builder, len(self.output))
            for i in reversed(range(len(self.output))):
                builder.PrependUOffsetTRelative(outputlist[i])
            output = builder.EndVector()
        if self.name is not None:
            name = builder.CreateString(self.name)
        if self.opType is not None:
            opType = builder.CreateString(self.opType)
        if self.domain is not None:
            domain = builder.CreateString(self.domain)
        if self.attribute is not None:
            attributelist = []
            for i in range(len(self.attribute)):
                attributelist.append(self.attribute[i].Pack(builder))
            NodeStartAttributeVector(builder, len(self.attribute))
            for i in reversed(range(len(self.attribute))):
                builder.PrependUOffsetTRelative(attributelist[i])
            attribute = builder.EndVector()
        if self.docString is not None:
            docString = builder.CreateString(self.docString)
        NodeStart(builder)
        if self.input is not None:
            NodeAddInput(builder, input)
        if self.output is not None:
            NodeAddOutput(builder, output)
        if self.name is not None:
            NodeAddName(builder, name)
        if self.opType is not None:
            NodeAddOpType(builder, opType)
        if self.domain is not None:
            NodeAddDomain(builder, domain)
        if self.attribute is not None:
            NodeAddAttribute(builder, attribute)
        if self.docString is not None:
            NodeAddDocString(builder, docString)
        node = NodeEnd(builder)
        return node



from typing import List

from esp_ppq.IR import BaseGraph, GraphFormatter


def truncate_graph(graph: BaseGraph, outputs: List[str]):
    """truncate your graph, so that all operations behind outputs(function
    parameter) will be eliminated. A list of output variable is given as
    parameter of this function. PPQ will goes forward from all those variables,
    mark all downstream operations for removing.

    A truncated graph object will return as result.

    ATTENTION: do not attempt to delete input variable.
    ATTETNION: you should invoke this function before quantization.

    Args:
        graph (BaseGraph): graph to be truncated
        outputs (List[str]): truncating from where

    Raises:
        KeyError: truncating variable is not in graph

    Returns:
        [type]: truncated graph
    """
    for output in outputs:
        if output not in graph.variables:
            raise KeyError(f'Can not find variable {output} in current graph.')
    processor = GraphFormatter(graph)

    for output in outputs:
        output_var = graph.variables[output]
        processor.truncate_on_var(output_var, mark_as_output=True)
    processor.delete_isolated()
    return graph



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class TensorAnnotation(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TensorAnnotation()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTensorAnnotation(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # TensorAnnotation
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # TensorAnnotation
    def TensorName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # TensorAnnotation
    def QuantParameterTensorNames(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.StringStringEntry import StringStringEntry
            obj = StringStringEntry()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # TensorAnnotation
    def QuantParameterTensorNamesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TensorAnnotation
    def QuantParameterTensorNamesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

def TensorAnnotationStart(builder):
    builder.StartObject(2)

def Start(builder):
    TensorAnnotationStart(builder)

def TensorAnnotationAddTensorName(builder, tensorName):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(tensorName), 0)

def AddTensorName(builder, tensorName):
    TensorAnnotationAddTensorName(builder, tensorName)

def TensorAnnotationAddQuantParameterTensorNames(builder, quantParameterTensorNames):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(quantParameterTensorNames), 0)

def AddQuantParameterTensorNames(builder, quantParameterTensorNames):
    TensorAnnotationAddQuantParameterTensorNames(builder, quantParameterTensorNames)

def TensorAnnotationStartQuantParameterTensorNamesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartQuantParameterTensorNamesVector(builder, numElems):
    return TensorAnnotationStartQuantParameterTensorNamesVector(builder, numElems)

def TensorAnnotationEnd(builder):
    return builder.EndObject()

def End(builder):
    return TensorAnnotationEnd(builder)

import FlatBuffers.Dl.StringStringEntry
try:
    from typing import List
except:
    pass

class TensorAnnotationT(object):

    # TensorAnnotationT
    def __init__(self):
        self.tensorName = None  # type: str
        self.quantParameterTensorNames = None  # type: List[FlatBuffers.Dl.StringStringEntry.StringStringEntryT]

    @classmethod
    def InitFromBuf(cls, buf, pos):
        tensorAnnotation = TensorAnnotation()
        tensorAnnotation.Init(buf, pos)
        return cls.InitFromObj(tensorAnnotation)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, tensorAnnotation):
        x = TensorAnnotationT()
        x._UnPack(tensorAnnotation)
        return x

    # TensorAnnotationT
    def _UnPack(self, tensorAnnotation):
        if tensorAnnotation is None:
            return
        self.tensorName = tensorAnnotation.TensorName()
        if not tensorAnnotation.QuantParameterTensorNamesIsNone():
            self.quantParameterTensorNames = []
            for i in range(tensorAnnotation.QuantParameterTensorNamesLength()):
                if tensorAnnotation.QuantParameterTensorNames(i) is None:
                    self.quantParameterTensorNames.append(None)
                else:
                    stringStringEntry_ = FlatBuffers.Dl.StringStringEntry.StringStringEntryT.InitFromObj(tensorAnnotation.QuantParameterTensorNames(i))
                    self.quantParameterTensorNames.append(stringStringEntry_)

    # TensorAnnotationT
    def Pack(self, builder):
        if self.tensorName is not None:
            tensorName = builder.CreateString(self.tensorName)
        if self.quantParameterTensorNames is not None:
            quantParameterTensorNameslist = []
            for i in range(len(self.quantParameterTensorNames)):
                quantParameterTensorNameslist.append(self.quantParameterTensorNames[i].Pack(builder))
            TensorAnnotationStartQuantParameterTensorNamesVector(builder, len(self.quantParameterTensorNames))
            for i in reversed(range(len(self.quantParameterTensorNames))):
                builder.PrependUOffsetTRelative(quantParameterTensorNameslist[i])
            quantParameterTensorNames = builder.EndVector()
        TensorAnnotationStart(builder)
        if self.tensorName is not None:
            TensorAnnotationAddTensorName(builder, tensorName)
        if self.quantParameterTensorNames is not None:
            TensorAnnotationAddQuantParameterTensorNames(builder, quantParameterTensorNames)
        tensorAnnotation = TensorAnnotationEnd(builder)
        return tensorAnnotation
